"""`#include <linux/fuse.h>`

FUSE doesn't use unique prefixes for each set of flags. All the FUSE flags start with just "FUSE_"
instead of starting with, e.g., "FUSE_INIT_" for the flags used in the initialization message,
"FUSE_WRITE_" for the flags used in the write message, etc..

To follow this pattern in rsyscall, so that e.g. FUSE_ASYNC_READ was used as FUSE.ASYNC_READ, would
require one big enum for all the flags used in all the messages. This is not very type-safe, so we
make up separate enums for each set of flags. So FUSE_ASYNC_READ is FUSE_INIT.ASYNC_READ, etc..

"""
from __future__ import annotations
import abc
import enum
from dataclasses import dataclass
from rsyscall._raw import lib, ffi # type: ignore
from rsyscall.struct import Struct, Serializable, HasSerializer, Serializer
from rsyscall.unistd import O
from rsyscall.time import Timespec
from rsyscall.sys.stat import TypeMode
from rsyscall.linux.dirent import DT
from rsyscall.memory.allocator import align
from rsyscall import Path
import os
import typing as t


class FUSE_INIT(enum.IntFlag):
    NONE = 0
    ASYNC_READ = lib.FUSE_ASYNC_READ
    POSIX_LOCKS = lib.FUSE_POSIX_LOCKS
    FILE_OPS = lib.FUSE_FILE_OPS
    ATOMIC_O_TRUNC = lib.FUSE_ATOMIC_O_TRUNC
    EXPORT_SUPPORT = lib.FUSE_EXPORT_SUPPORT
    BIG_WRITES = lib.FUSE_BIG_WRITES
    DONT_MASK = lib.FUSE_DONT_MASK
    SPLICE_WRITE = lib.FUSE_SPLICE_WRITE
    SPLICE_MOVE = lib.FUSE_SPLICE_MOVE
    SPLICE_READ = lib.FUSE_SPLICE_READ
    FLOCK_LOCKS = lib.FUSE_FLOCK_LOCKS
    HAS_IOCTL_DIR = lib.FUSE_HAS_IOCTL_DIR
    AUTO_INVAL_DATA = lib.FUSE_AUTO_INVAL_DATA
    DO_READDIRPLUS = lib.FUSE_DO_READDIRPLUS
    READDIRPLUS_AUTO = lib.FUSE_READDIRPLUS_AUTO
    ASYNC_DIO = lib.FUSE_ASYNC_DIO
    WRITEBACK_CACHE = lib.FUSE_WRITEBACK_CACHE
    NO_OPEN_SUPPORT = lib.FUSE_NO_OPEN_SUPPORT
    PARALLEL_DIROPS = lib.FUSE_PARALLEL_DIROPS
    HANDLE_KILLPRIV = lib.FUSE_HANDLE_KILLPRIV
    POSIX_ACL = lib.FUSE_POSIX_ACL
    ABORT_ERROR = lib.FUSE_ABORT_ERROR

class FUSE_WRITE(enum.IntFlag):
    CACHE = lib.FUSE_WRITE_CACHE
    LOCKOWNER = lib.FUSE_WRITE_LOCKOWNER

class FUSE_RELEASE(enum.IntFlag):
    FLUSH = lib.FUSE_RELEASE_FLUSH
    FLOCK_UNLOCK = lib.FUSE_RELEASE_FLOCK_UNLOCK

class FUSE_OPCODE(enum.IntEnum):
    LOOKUP = lib.FUSE_LOOKUP
    FORGET = lib.FUSE_FORGET
    GETATTR = lib.FUSE_GETATTR
    SETATTR = lib.FUSE_SETATTR
    READLINK = lib.FUSE_READLINK
    SYMLINK = lib.FUSE_SYMLINK
    MKNOD = lib.FUSE_MKNOD
    MKDIR = lib.FUSE_MKDIR
    UNLINK = lib.FUSE_UNLINK
    RMDIR = lib.FUSE_RMDIR
    RENAME = lib.FUSE_RENAME
    LINK = lib.FUSE_LINK
    OPEN = lib.FUSE_OPEN
    READ = lib.FUSE_READ
    WRITE = lib.FUSE_WRITE
    STATFS = lib.FUSE_STATFS
    RELEASE = lib.FUSE_RELEASE
    FSYNC = lib.FUSE_FSYNC
    SETXATTR = lib.FUSE_SETXATTR
    GETXATTR = lib.FUSE_GETXATTR
    LISTXATTR = lib.FUSE_LISTXATTR
    REMOVEXATTR = lib.FUSE_REMOVEXATTR
    FLUSH = lib.FUSE_FLUSH
    INIT = lib.FUSE_INIT
    OPENDIR = lib.FUSE_OPENDIR
    READDIR = lib.FUSE_READDIR
    RELEASEDIR = lib.FUSE_RELEASEDIR
    FSYNCDIR = lib.FUSE_FSYNCDIR
    GETLK = lib.FUSE_GETLK
    SETLK = lib.FUSE_SETLK
    SETLKW = lib.FUSE_SETLKW
    ACCESS = lib.FUSE_ACCESS
    CREATE = lib.FUSE_CREATE
    INTERRUPT = lib.FUSE_INTERRUPT
    BMAP = lib.FUSE_BMAP
    DESTROY = lib.FUSE_DESTROY
    IOCTL = lib.FUSE_IOCTL
    POLL = lib.FUSE_POLL
    NOTIFY_REPLY = lib.FUSE_NOTIFY_REPLY
    BATCH_FORGET = lib.FUSE_BATCH_FORGET
    FALLOCATE = lib.FUSE_FALLOCATE
    READDIRPLUS = lib.FUSE_READDIRPLUS
    RENAME2 = lib.FUSE_RENAME2
    LSEEK = lib.FUSE_LSEEK

class FOPEN(enum.IntEnum):
    NONE = 0
    DIRECT_IO = lib.FOPEN_DIRECT_IO
    KEEP_CACHE = lib.FOPEN_KEEP_CACHE
    NONSEEKABLE = lib.FOPEN_NONSEEKABLE

class FUSE_ATTR(enum.IntEnum):
    NONE = 0
    SUBMOUNT = lib.FUSE_ATTR_SUBMOUNT

class FUSE_OPEN(enum.IntEnum):
    NONE = 0
    KILL_SUIDGID = lib.FUSE_OPEN_KILL_SUIDGID

FUSE_MIN_READ_BUFFER = lib.FUSE_MIN_READ_BUFFER

@dataclass
class FuseAttr(Struct):
    ino: int
    size: int
    blocks: int
    atime: Timespec
    mtime: Timespec
    ctime: Timespec
    mode: TypeMode
    nlink: int
    uid: int
    gid: int
    rdev: int
    blksize: int
    flags: FUSE_ATTR=FUSE_ATTR.NONE

    def _to_cffi_dict(self) -> t.Dict[str, int]:
        return {
	    "ino": self.ino,
	    "size": self.size,
	    "blocks": self.blocks,
	    "atime": self.atime.sec,
	    "mtime": self.mtime.sec,
	    "ctime": self.ctime.sec,
	    "atimensec": self.atime.nsec,
	    "mtimensec": self.mtime.nsec,
	    "ctimensec": self.ctime.nsec,
	    "mode": int(self.mode),
	    "nlink": self.nlink,
	    "uid": self.uid,
	    "gid": self.gid,
	    "rdev": self.rdev,
	    "blksize": self.blksize,
            "flags": self.flags,
        }

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_attr const*', self._to_cffi_dict())))

    T = t.TypeVar('T', bound='FuseAttr')
    @classmethod
    def from_cffi(cls: t.Type[T], struct) -> T:
        return cls(
	    ino=struct.ino,
	    size=struct.size,
	    blocks=struct.blocks,
	    atime=Timespec(sec=struct.atime, nsec=struct.atimensec),
	    mtime=Timespec(sec=struct.mtime, nsec=struct.mtimensec),
	    ctime=Timespec(sec=struct.ctime, nsec=struct.ctimensec),
	    mode=struct.mode,
	    nlink=struct.nlink,
	    uid=struct.uid,
	    gid=struct.gid,
	    rdev=struct.rdev,
	    blksize=struct.blksize,
	    flags=FUSE_ATTR(struct.flags),
        )

    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        return cls.from_cffi(ffi.cast('struct fuse_attr*', ffi.from_buffer(data)))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_attr')

@dataclass
class FuseInHeader:
    unique: int
    # nodeid 1 is the root of the filesystem; all other nodeids are returned by the FUSE server
    nodeid: int
    uid: int
    gid: int
    pid: int

    @staticmethod
    def from_cffi(struct) -> FuseInHeader:
        return FuseInHeader(
            unique=struct.unique,
            nodeid=struct.nodeid,
            uid=struct.uid,
            gid=struct.gid,
            pid=struct.pid,
        )

    @staticmethod
    def sizeof() -> int:
        return ffi.sizeof('struct fuse_in_header')

# mypy seems to not like dataclasses with abstract methods
@dataclass # type: ignore
class FuseIn(Serializable):
    hdr: FuseInHeader
    opcode: t.ClassVar[FUSE_OPCODE]

    T = t.TypeVar('T', bound='FuseIn')
    @classmethod
    @abc.abstractmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T: ...

    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_in_header*', ffi.from_buffer(data))
        if struct.len > len(data):
            raise Exception("only part of the FUSE in packet was passed to from_bytes",
                            "length field in header is", struct.len, "but passed buffer is only", len(data))
        opcode = FUSE_OPCODE(struct.opcode)
        if opcode != cls.opcode:
            raise Exception("mismatch between opcode in header", opcode, "and opcode for this class", cls.opcode)
        hdr = FuseInHeader.from_cffi(struct)
        variable = data[FuseInHeader.sizeof():struct.len]
        msg = cls.from_header(hdr, variable)
        return msg
    
    @abc.abstractmethod
    def msg_to_bytes(self) -> bytes: ...

    def to_bytes(self) -> bytes:
        msg_data = self.msg_to_bytes()
        return bytes(ffi.buffer(ffi.new('struct fuse_in_header*', {
            'len': FuseInHeader.sizeof() + len(msg_data),
            'opcode': self.opcode,
            'unique': self.hdr.unique,
            'nodeid': self.hdr.nodeid,
            'uid': self.hdr.uid,
            'gid': self.hdr.gid,
            'pid': self.hdr.pid,
        }))) + msg_data

@dataclass
class FuseInitIn(Struct):
    major: int
    minor: int
    max_readahead: int
    flags: FUSE_INIT
    opcode = FUSE_OPCODE.INIT

    T = t.TypeVar('T', bound='FuseInitIn')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_init_in*', ffi.from_buffer(data))
        value = cls(
            major=struct.major,
            minor=struct.minor,
            max_readahead=struct.max_readahead,
            flags=FUSE_INIT(struct.flags),
        )
        return value

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_init_in*', {
            'major': self.major,
            'minor': self.minor,
            'max_readahead': self.max_readahead,
            'flags': self.flags,
        })))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_init_in')

@dataclass
class FuseInitOp(FuseIn, Struct):
    msg: FuseInitIn
    opcode = FUSE_OPCODE.INIT

    T = t.TypeVar('T', bound='FuseInitOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseInitIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseInitIn.sizeof()

    def respond(self, msg: FuseInitOut, error: int=0) -> FuseInitResponse:
        return FuseInitResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=msg)

@dataclass
class FuseLookupOp(FuseIn):
    name: str
    opcode = FUSE_OPCODE.LOOKUP

    T = t.TypeVar('T', bound='FuseLookupOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(
            hdr=hdr,
            # strip trailing null byte
            name=data[:-1].decode(),
        )

    def msg_to_bytes(self) -> bytes:
        return self.name.encode() + b'\0'

    def respond(self, msg: FuseEntryOut, error: int=0) -> FuseLookupResponse:
        return FuseLookupResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=msg)

@dataclass
class FuseOpenIn(Struct):
    flags: O
    open_flags: FUSE_OPEN=FUSE_OPEN.NONE

    T = t.TypeVar('T', bound='FuseOpenIn')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_open_in*', ffi.from_buffer(data))
        return cls(flags=O(struct.flags), open_flags=FUSE_OPEN(struct.open_flags))

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_open_in*', {
            'flags': self.flags,
            'open_flags': self.open_flags,
        })))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_open_in')

@dataclass
class FuseOpenOp(FuseIn, Struct):
    msg: FuseOpenIn
    opcode = FUSE_OPCODE.OPEN

    T = t.TypeVar('T', bound='FuseOpenOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseOpenIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseOpenIn.sizeof()

    def respond(self, msg: FuseOpenOut, error: int=0) -> FuseOpenResponse:
        return FuseOpenResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=msg)

@dataclass
class FuseOpendirOp(FuseIn, Struct):
    msg: FuseOpenIn
    opcode = FUSE_OPCODE.OPENDIR

    T = t.TypeVar('T', bound='FuseOpendirOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseOpenIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseOpenIn.sizeof()

    def respond(self, msg: FuseOpenOut, error: int=0) -> FuseOpendirResponse:
        return FuseOpendirResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=msg)

class FUSE_READ(enum.IntFlag):
    NONE = 0
    LOCKOWNER = lib.FUSE_READ_LOCKOWNER

@dataclass
class FuseReadIn(Struct):
    fh: int
    offset: int
    size: int
    read_flags: FUSE_READ
    # the O flags currently set on the file; these can be changed after open with fcntl
    flags: O

    T = t.TypeVar('T', bound='FuseReadIn')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_read_in*', ffi.from_buffer(data))
        return cls(
            fh=struct.fh,
            offset=struct.offset,
            size=struct.size,
            read_flags=FUSE_READ(struct.read_flags),
            flags=O(struct.flags),
        )

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_read_in*', {
            'fh': self.fh,
            'offset': self.offset,
            'size': self.size,
            'read_flags': self.read_flags,
            'flags': self.flags,
            'padding': 0,
        })))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_read_in')

@dataclass
class FuseReadOp(FuseIn, Struct):
    msg: FuseReadIn
    opcode = FUSE_OPCODE.READ

    T = t.TypeVar('T', bound='FuseReadOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseReadIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseReadIn.sizeof()

    def respond(self, msg: bytes, error: int=0) -> FuseReadResponse:
        return FuseReadResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=msg)

@dataclass
class FuseReaddirOp(FuseIn, Struct):
    msg: FuseReadIn
    opcode = FUSE_OPCODE.READDIR

    T = t.TypeVar('T', bound='FuseReaddirOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseReadIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseReadIn.sizeof()

    def respond(self, msg: t.List[FuseDirent], error: int=0) -> FuseReaddirResponse:
        return FuseReaddirResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=msg)

@dataclass
class FuseReaddirplusOp(FuseIn, Struct):
    msg: FuseReadIn
    opcode = FUSE_OPCODE.READDIRPLUS

    T = t.TypeVar('T', bound='FuseReaddirplusOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseReadIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseReadIn.sizeof()

    def respond(self, msg: t.List[FuseDirentplus], error: int=0) -> FuseReaddirplusResponse:
        return FuseReaddirplusResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=msg)

@dataclass
class FuseReadlinkOp(FuseIn, Struct):
    opcode = FUSE_OPCODE.READLINK

    T = t.TypeVar('T', bound='FuseReadlinkOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr)

    def msg_to_bytes(self) -> bytes:
        return b""

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof()

    def respond(self, msg: t.Union[str, os.PathLike], error: int=0) -> FuseReadlinkResponse:
        return FuseReadlinkResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=msg)

class FUSE_GETATTR(enum.IntFlag):
    NONE = 0
    FH = lib.FUSE_GETATTR_FH

@dataclass
class FuseGetattrIn(Struct):
    getattr_flags: FUSE_GETATTR
    fh: int

    T = t.TypeVar('T', bound='FuseGetattrIn')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_getattr_in*', ffi.from_buffer(data))
        return cls(
            getattr_flags=FUSE_GETATTR(struct.getattr_flags),
            fh=struct.fh,
        )

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_getattr_in*', {
            'getattr_flags': self.getattr_flags,
            'dummy': 0,
            'fh': self.fh,
        })))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_getattr_in')

@dataclass
class FuseFlushIn(Struct):
    fh: int
    lock_owner: int

    T = t.TypeVar('T', bound='FuseFlushIn')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_flush_in*', ffi.from_buffer(data))
        return cls(fh=struct.fh, lock_owner=struct.lock_owner)

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_flush_in*', {
            'fh': self.fh,
            'unused': 0,
            'padding': 0,
            'lock_owner': self.lock_owner,
        })))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_flush_in')

@dataclass
class FuseFlushOp(FuseIn, Struct):
    msg: FuseFlushIn
    opcode = FUSE_OPCODE.FLUSH

    T = t.TypeVar('T', bound='FuseFlushOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseFlushIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseFlushIn.sizeof()

    def respond(self, error: int=0) -> FuseFlushResponse:
        return FuseFlushResponse(FuseOutHeader(error=error, unique=self.hdr.unique))

@dataclass
class FuseReleaseIn(Struct):
    fh: int
    flags: O
    release_flags: FUSE_RELEASE
    lock_owner: int

    T = t.TypeVar('T', bound='FuseReleaseIn')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_release_in*', ffi.from_buffer(data))
        return cls(
            fh=struct.fh,
            flags=O(struct.flags),
            release_flags=FUSE_RELEASE(struct.release_flags),
            lock_owner=struct.lock_owner,
        )

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_release_in*', {
            'fh': self.fh,
            'flags': self.flags,
            'release_flags': self.release_flags,
            'unused': 0,
        })))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_release_in')

@dataclass
class FuseReleaseOp(FuseIn, Struct):
    msg: FuseReleaseIn
    opcode = FUSE_OPCODE.RELEASE

    T = t.TypeVar('T', bound='FuseReleaseOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseReleaseIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseReleaseIn.sizeof()

    def respond(self, error: int=0) -> FuseReleaseResponse:
        return FuseReleaseResponse(FuseOutHeader(error=error, unique=self.hdr.unique))

@dataclass
class FuseReleasedirOp(FuseIn, Struct):
    msg: FuseReleaseIn
    opcode = FUSE_OPCODE.RELEASEDIR

    T = t.TypeVar('T', bound='FuseReleasedirOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseReleaseIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseReleaseIn.sizeof()

    def respond(self, error: int=0) -> FuseReleasedirResponse:
        return FuseReleasedirResponse(FuseOutHeader(error=error, unique=self.hdr.unique))

@dataclass
class FuseGetattrOp(FuseIn, Struct):
    msg: FuseGetattrIn
    opcode = FUSE_OPCODE.GETATTR

    T = t.TypeVar('T', bound='FuseGetattrOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseGetattrIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @classmethod
    def sizeof(cls) -> int:
        return FuseInHeader.sizeof() + FuseGetattrIn.sizeof()

    def respond(self, msg: FuseAttrOut, error: int=0) -> FuseGetattrResponse:
        return FuseGetattrResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=msg)

@dataclass
class FuseGetxattrIn(Serializable):
    data: str

    T = t.TypeVar('T', bound='FuseGetxattrIn')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_getxattr_in*', ffi.from_buffer(data))
        varsize = struct.size - 4
        variable = data[ffi.sizeof('struct fuse_getxattr_in'):]
        if len(variable) < varsize:
            raise Exception("partial fuse_getxattr_in packet, received data", len(variable), variable,
                            "expected size", struct.size)
        # -1 to strip null byte
        return cls(os.fsdecode(variable[:varsize-1]))

    def to_bytes(self) -> bytes:
        data = self.data + "\0"
        return bytes(ffi.buffer(ffi.new('struct fuse_getxattr_in*', {
            'size': len(data),
            'padding': 0,
        })))

@dataclass
class FuseGetxattrOp(FuseIn):
    # aha ok we get the name, which we have the size of.
    # that includes the null terminator
    # then we respond with the xattr data presumably
    msg: FuseGetxattrIn
    opcode = FUSE_OPCODE.GETXATTR

    T = t.TypeVar('T', bound='FuseGetxattrOp')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseInHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseGetxattrIn.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    def respond(self, data: str, error: int=0) -> FuseGetxattrResponse:
        return FuseGetxattrResponse(FuseOutHeader(error=error, unique=self.hdr.unique), msg=FuseGetxattrOut(data))

    def error(self, error: int) -> FuseErrorResponse:
        return FuseErrorResponse(FuseOutHeader(error=error, unique=self.hdr.unique))

opcode_classes: t.Dict[FUSE_OPCODE, t.Type[FuseIn]] = {
    FUSE_OPCODE.INIT: FuseInitOp,
    FUSE_OPCODE.LOOKUP: FuseLookupOp,
    FUSE_OPCODE.OPEN: FuseOpenOp,
    FUSE_OPCODE.OPENDIR: FuseOpendirOp,
    FUSE_OPCODE.READ: FuseReadOp,
    FUSE_OPCODE.READDIR: FuseReaddirOp,
    FUSE_OPCODE.READDIRPLUS: FuseReaddirplusOp,
    FUSE_OPCODE.READLINK: FuseReadlinkOp,
    FUSE_OPCODE.GETATTR: FuseGetattrOp,
    FUSE_OPCODE.GETXATTR: FuseGetxattrOp,
    FUSE_OPCODE.FLUSH: FuseFlushOp,
    FUSE_OPCODE.RELEASE: FuseReleaseOp,
    FUSE_OPCODE.RELEASEDIR: FuseReleasedirOp,
}

def fuse_in_parse_split(data: bytes) -> t.Tuple[FuseIn, bytes]:
    "Split the buffer into parsed header with variable length section, and remaining unparsed data"
    struct = ffi.cast('struct fuse_in_header*', ffi.from_buffer(data))
    if struct.len > len(data):
        raise Exception("only part of the FUSE packet was passed to _header_parse_split",
                        "length field in header is", struct.len, "but passed buffer is only", len(data))
    opcode = FUSE_OPCODE(struct.opcode)
    hdr = FuseInHeader.from_cffi(struct)
    variable = data[FuseInHeader.sizeof():struct.len]
    rest = data[struct.len:]
    # dispatch to the specific message type
    msg = opcode_classes[opcode].from_header(hdr, variable)
    return msg, rest

class FuseInList(t.List[FuseIn], Serializable):
    def to_bytes(self) -> bytes:
        ret = b""
        for ent in self:
            ret += ent.to_bytes()
        return ret

    T = t.TypeVar('T', bound='FuseInList')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        entries = []
        while data:
            # the FUSE fd will never give us a partial packet, so we don't have to rebuffer
            pkt, data = fuse_in_parse_split(data)
            entries.append(pkt)
        return cls(entries)

@dataclass
class FuseOutHeader:
    error: int
    unique: int

    @staticmethod
    def from_cffi(struct) -> FuseOutHeader:
        return FuseOutHeader(
            error=struct.error,
            unique=struct.unique,
        )

    @staticmethod
    def sizeof() -> int:
        return ffi.sizeof('struct fuse_out_header')

# mypy seems to not like dataclasses with abstract methods
@dataclass # type: ignore
class FuseOut(Serializable):
    hdr: FuseOutHeader

    T = t.TypeVar('T', bound='FuseOut')
    # @abc.abstractclassmethod
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T: ...

    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_out_header*', ffi.from_buffer(data))
        if struct.len > len(data):
            raise Exception("only part of the FUSE out packet was passed to from_bytes",
                            "length field in header is", struct.len, "but passed buffer is only", len(data))
        msg_data = data[ffi.sizeof('struct fuse_out_header'):struct.len]
        return cls.from_header(FuseOutHeader.from_cffi(struct), msg_data)
    
    @abc.abstractmethod
    def msg_to_bytes(self) -> bytes: ...

    def to_bytes(self) -> bytes:
        msg_data = self.msg_to_bytes()
        return bytes(ffi.buffer(ffi.new('struct fuse_out_header*', {
            'len': ffi.sizeof('struct fuse_out_header') + len(msg_data),
            'error': self.hdr.error,
            'unique': self.hdr.unique,
        }))) + msg_data

@dataclass
class FuseErrorResponse(FuseOut, Struct):
    T = t.TypeVar('T', bound='FuseErrorResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr)

    def msg_to_bytes(self) -> bytes:
        return b""

    @staticmethod
    def sizeof() -> int:
        return FuseOutHeader.sizeof()

@dataclass
class FuseInitOut(Struct):
    major: int
    minor: int
    max_readahead: int
    flags: FUSE_INIT
    max_background: int
    congestion_threshold: int
    max_write: int
    time_gran: int

    T = t.TypeVar('T', bound='FuseInitOut')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_init_out*', ffi.from_buffer(data))
        return cls(
            major=struct.major,
            minor=struct.minor,
            max_readahead=struct.max_readahead,
            flags=FUSE_INIT(struct.flags),
            max_background=struct.max_background,
            congestion_threshold=struct.congestion_threshold,
            max_write=struct.max_write,
            time_gran=struct.time_gran,
        )

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_init_out*', {
            'major': self.major,
            'minor': self.minor,
            'max_readahead': self.max_readahead,
            'flags': self.flags,
            'max_background': self.max_background,
            'congestion_threshold': self.congestion_threshold,
            'max_write': self.max_write,
            'time_gran': self.time_gran,
        })))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_init_out')

@dataclass
class FuseInitResponse(FuseOut, Struct):
    msg: FuseInitOut

    T = t.TypeVar('T', bound='FuseInitResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseInitOut.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @staticmethod
    def sizeof() -> int:
        return FuseOutHeader.sizeof() + FuseInitOut.sizeof()

@dataclass
class FuseEntryOut(Struct):
    nodeid: int
    generation: int
    entry_valid: Timespec
    attr_valid: Timespec
    attr: FuseAttr

    T = t.TypeVar('T', bound='FuseEntryOut')
    @classmethod
    def from_cffi(cls: t.Type[T], struct) -> T:
        return cls(
            nodeid=struct.nodeid,
            generation=struct.generation,
            entry_valid=Timespec(struct.entry_valid, struct.entry_valid_nsec),
            attr_valid=Timespec(struct.attr_valid, struct.attr_valid_nsec),
            attr=FuseAttr.from_cffi(struct.attr),
        )

    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        return cls.from_cffi(ffi.cast('struct fuse_entry_out*', ffi.from_buffer(data)))

    def _to_cffi_dict(self) -> t.Dict[str, t.Any]:
        return {
            'nodeid': self.nodeid,
            'generation': self.generation,
            'entry_valid': self.entry_valid.sec,
            'attr_valid': self.attr_valid.sec,
            'entry_valid_nsec': self.entry_valid.nsec,
            'attr_valid_nsec': self.attr_valid.nsec,
            'attr': self.attr._to_cffi_dict(),
        }

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_entry_out*', self._to_cffi_dict())))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_entry_out')

@dataclass
class FuseLookupResponse(FuseOut, Struct):
    msg: FuseEntryOut

    T = t.TypeVar('T', bound='FuseLookupResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseEntryOut.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @staticmethod
    def sizeof() -> int:
        return FuseOutHeader.sizeof() + FuseEntryOut.sizeof()

@dataclass
class FuseOpenOut(Struct):
    fh: int
    open_flags: FOPEN

    T = t.TypeVar('T', bound='FuseOpenOut')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_open_out*', ffi.from_buffer(data))
        return cls(
            fh=struct.fh,
            open_flags=FOPEN(struct.open_flags),
        )

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_open_out*', {
            'fh': self.fh,
            'open_flags': self.open_flags,
            'padding': 0,
        })))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_open_out')

@dataclass
class FuseOpenResponse(FuseOut, Struct):
    msg: FuseOpenOut

    T = t.TypeVar('T', bound='FuseOpenResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseOpenOut.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @staticmethod
    def sizeof() -> int:
        return FuseOutHeader.sizeof() + FuseOpenOut.sizeof()

@dataclass
class FuseOpendirResponse(FuseOut, Struct):
    msg: FuseOpenOut

    T = t.TypeVar('T', bound='FuseOpendirResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseOpenOut.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @staticmethod
    def sizeof() -> int:
        return FuseOutHeader.sizeof() + FuseOpenOut.sizeof()

@dataclass
class FuseReadResponse(FuseOut):
    msg: bytes

    T = t.TypeVar('T', bound='FuseReadResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=data)

    def msg_to_bytes(self) -> bytes:
        return self.msg

@dataclass
class FuseAttrOut(Struct):
    attr_valid: Timespec
    attr: FuseAttr

    T = t.TypeVar('T', bound='FuseAttrOut')
    @classmethod
    def from_header(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_attr_out*', ffi.from_buffer(data))
        return cls(
            attr_valid=Timespec(struct.attr_valid, struct.attr_valid_nsec),
            attr=FuseAttr.from_cffi(struct.attr),
        )

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct fuse_attr_out*', {
            'attr_valid': self.attr_valid.sec,
            'attr_valid_nsec': self.attr_valid.nsec,
            'dummy': 0,
            'attr': self.attr._to_cffi_dict(),
        })))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fuse_attr_out')

@dataclass
class FuseGetattrResponse(FuseOut, Struct):
    msg: FuseAttrOut

    T = t.TypeVar('T', bound='FuseGetattrResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseAttrOut.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()

    @staticmethod
    def sizeof() -> int:
        return FuseOutHeader.sizeof() + FuseAttrOut.sizeof()

@dataclass
class FuseDirent:
    ino: int
    off: int # the offset to seek to to see the next dirent
    type: DT
    name: str

    def _to_cffi_dict(self, name: bytes) -> t.Dict[str, t.Any]:
        return {
            "ino": self.ino,
            "off": self.off,
            "namelen": len(name),
            "type": self.type,
            "name": name,
        }

    @classmethod
    def from_cffi(cls, struct) -> FuseDirent:
        return cls(
            ino=struct.ino,
            off=struct.off,
            type=DT(struct.type),
            # the name is padded with null bytes to make the dirent aligned,
            # so we have to use strlen to find the end
            name=ffi.string(struct.name, struct.namelen).decode(),
        )

    def to_bytes(self) -> bytes:
        namebytes = self.name.encode() + b'\0'
        length = ffi.sizeof('struct fuse_dirent') + len(namebytes)
        padding = align(length, 8) - length
        namebytes = namebytes + bytes(padding)
        return bytes(ffi.buffer(ffi.new('struct fuse_dirent*',
                                        self._to_cffi_dict(namebytes)
        ))) + namebytes

@dataclass
class FuseReaddirResponse(FuseOut):
    msg: t.List[FuseDirent]

    T = t.TypeVar('T', bound='FuseReaddirResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        entries = []
        while len(data) > 0:
            # We do the work of from_bytes in this class instead of in Dirent because we need the
            # raw length field from the struct; merely doing len(name) will exclude padding.
            struct = ffi.cast('struct fuse_dirent*', ffi.from_buffer(data))
            record_length = ffi.sizeof('struct fuse_dirent') + struct.namelen
            if len(data) < record_length:
                raise Exception("partial packet passed to FuseDirent.from_bytes")
            entries.append(FuseDirent.from_cffi(struct))
            data = data[record_length:]
        return cls(hdr=hdr, msg=entries)

    def msg_to_bytes(self) -> bytes:
        ret = b""
        for ent in self.msg:
            ret += ent.to_bytes()
        return ret

@dataclass
class FuseDirentplus:
    entry_out: FuseEntryOut
    dirent: FuseDirent

    @classmethod
    def from_cffi(cls, struct) -> FuseDirentplus:
        return cls(
            entry_out=FuseEntryOut.from_cffi(struct.entry_out),
            dirent=FuseDirent.from_cffi(struct.dirent),
        )

    def to_bytes(self) -> bytes:
        namebytes = self.dirent.name.encode() + b'\0'
        length = ffi.sizeof('struct fuse_direntplus') + len(namebytes)
        padding = align(length, 8) - length
        namebytes = namebytes + bytes(padding)
        return bytes(ffi.buffer(ffi.new('struct fuse_direntplus*', {
            "entry_out": self.entry_out._to_cffi_dict(),
            "dirent": self.dirent._to_cffi_dict(namebytes),
        }))) + namebytes

@dataclass
class FuseReaddirplusResponse(FuseOut):
    msg: t.List[FuseDirentplus]

    T = t.TypeVar('T', bound='FuseReaddirplusResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        entries = []
        while len(data) > 0:
            # We do the work of from_bytes in this class instead of in Direntplus because we need the
            # raw length field from the struct; merely doing len(name) will exclude padding.
            struct = ffi.cast('struct fuse_direntplus*', ffi.from_buffer(data))
            record_length = ffi.sizeof('struct fuse_direntplus') + struct.namelen
            if len(data) < record_length:
                raise Exception("partial packet passed to FuseDirentplus.from_bytes")
            entries.append(FuseDirentplus.from_cffi(struct))
            data = data[record_length:]
        return cls(hdr=hdr, msg=entries)

    def msg_to_bytes(self) -> bytes:
        ret = b""
        for ent in self.msg:
            ret += ent.to_bytes()
        return ret

@dataclass
class FuseReadlinkResponse(FuseOut):
    msg: t.Union[str, os.PathLike]

    T = t.TypeVar('T', bound='FuseReadlinkResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=os.fsdecode(data))

    def msg_to_bytes(self) -> bytes:
        return os.fsencode(self.msg)

@dataclass
class FuseFlushResponse(FuseOut, Struct):
    T = t.TypeVar('T', bound='FuseFlushResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr)

    def msg_to_bytes(self) -> bytes:
        return b""

    @staticmethod
    def sizeof() -> int:
        return FuseOutHeader.sizeof()

@dataclass
class FuseReleaseResponse(FuseOut, Struct):
    T = t.TypeVar('T', bound='FuseReleaseResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr)

    def msg_to_bytes(self) -> bytes:
        return b""

    @staticmethod
    def sizeof() -> int:
        return FuseOutHeader.sizeof()

@dataclass
class FuseReleasedirResponse(FuseOut, Struct):
    T = t.TypeVar('T', bound='FuseReleasedirResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr)

    def msg_to_bytes(self) -> bytes:
        return b""

    @staticmethod
    def sizeof() -> int:
        return FuseOutHeader.sizeof()


@dataclass
class FuseGetxattrOut(Serializable):
    data: str

    T = t.TypeVar('T', bound='FuseGetxattrOut')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct fuse_getxattr_out*', ffi.from_buffer(data))
        variable = data[ffi.sizeof('struct fuse_getxattr_out'):]
        if len(variable) < struct.size:
            raise Exception("partial fuse_getxattr_out packet, received size", len(variable),
                            "expected size", struct.size)
        # -1 to strip null byte
        return cls(os.fsdecode(variable[:struct.size-1]))

    def to_bytes(self) -> bytes:
        data = self.data + "\0"
        return bytes(ffi.buffer(ffi.new('struct fuse_getxattr_out*', {
            'size': len(data),
            'padding': 0,
        })))

@dataclass
class FuseGetxattrResponse(FuseOut):
    msg: FuseGetxattrOut

    T = t.TypeVar('T', bound='FuseGetxattrResponse')
    @classmethod
    def from_header(cls: t.Type[T], hdr: FuseOutHeader, data: bytes) -> T:
        return cls(hdr=hdr, msg=FuseGetxattrOut.from_bytes(data))

    def msg_to_bytes(self) -> bytes:
        return self.msg.to_bytes()


#### Tests ####
from unittest import TestCase
class TestFuse(TestCase):
    def test_fuse_in_list(self) -> None:
        initial = FuseInList([
            FuseInitOp(FuseInHeader(2, 3, 4, 5, 6), FuseInitIn(7, 8, 9, FUSE_INIT.BIG_WRITES|FUSE_INIT.DONT_MASK)),
        ])
        data = initial.to_bytes()
        output = FuseInList.from_bytes(data)
        self.assertEqual(initial, output)

    def test_fuse_init_out(self) -> None:
        initial = FuseInitResponse(FuseOutHeader(error=0, unique=1),
                                   FuseInitOut(2, 3, 4, FUSE_INIT.BIG_WRITES|FUSE_INIT.DONT_MASK, 6, 7, 8, 9))
        data = initial.to_bytes()
        output = FuseInitOut.from_bytes(data)
        self.assertEqual(initial, output)



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.sys.eventfd import *
from rsyscall.struct import Int64

class TestEventfd(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread
        self.fd = await self.thr.task.eventfd(0)

    async def asyncTearDown(self) -> None:
        await self.fd.close()

    async def test(self) -> None:
        inval = Int64(10)
        written, _ = await self.fd.write(await self.thr.ram.ptr(inval))
        read, _ = await self.fd.read(written)
        self.assertEqual(inval, await read.read())



"Our core interfaces for reading/writing memory."
from __future__ import annotations
import typing as t
import abc
if t.TYPE_CHECKING:
    from rsyscall.handle import Pointer, Task

class MemoryGateway:
    "This low-level class allows us to read bytes from memory and write bytes to memory."
    @abc.abstractmethod
    async def read(self, src: Pointer) -> bytes:
        "Read the memory pointed to by this Pointer, and return its contents as a bytestring."
        pass

    @abc.abstractmethod
    async def write(self, dest: Pointer, data: bytes) -> None:
        "Write this bytestring to the memory pointed to by this Pointer."
        pass

class MemoryTransport(MemoryGateway):
    @abc.abstractmethod
    def inherit(self, task: Task) -> MemoryTransport: ...



"`#include <sys/prctl.h>`"
from rsyscall._raw import lib # type: ignore
import typing as t
import enum

__all__ = [
    "PR",
    "PR_CAP_AMBIENT",
    "PrctlTask",
]

class PR(enum.IntEnum):
    GET_NAME = lib.PR_GET_NAME
    SET_NAME = lib.PR_SET_NAME
    SET_PDEATHSIG = lib.PR_SET_PDEATHSIG
    CAP_AMBIENT = lib.PR_CAP_AMBIENT

class PR_CAP_AMBIENT(enum.IntEnum):
    RAISE = lib.PR_CAP_AMBIENT_RAISE

#### Classes ####
import rsyscall.far
from rsyscall.handle.pointer import Pointer, WrittenPointer, ReadablePointer
from rsyscall.sys.capability import CAP
from rsyscall.signal import SIG

class PrctlTask(rsyscall.far.Task):
    "A base class providing the `prctl` syscall"

    async def prctl_set_pdeathsig(self, option: t.Literal[PR.SET_PDEATHSIG], arg2: t.Union[SIG, t.Literal[0]]) -> None:
        await _prctl(self.sysif, option, arg2)

    async def prctl_cap_ambient(self, option: t.Literal[PR.CAP_AMBIENT], arg2: PR_CAP_AMBIENT, arg3: CAP) -> int:
        return await _prctl(self.sysif, option, arg2, arg3)

    async def prctl_get_name(self, option: t.Literal[PR.GET_NAME], arg2: Pointer[str]) -> ReadablePointer[str]:
        with arg2.borrow(self) as arg2_n:
            await _prctl(self.sysif, option, arg2_n)
            return arg2.readable_split(16)[0]

    async def prctl_set_name(self, option: t.Literal[PR.SET_NAME], arg2: WrittenPointer[str]) -> None:
        with arg2.borrow(self) as arg2_n:
            await _prctl(self.sysif, option, arg2_n)
            return None

    @t.overload
    async def prctl(self, option: t.Literal[PR.SET_PDEATHSIG], arg2: t.Union[SIG, t.Literal[0]]) -> None: ...
    @t.overload
    async def prctl(self, option: t.Literal[PR.CAP_AMBIENT], arg2: PR_CAP_AMBIENT, arg3: CAP) -> int: ...
    @t.overload
    async def prctl(self, option: t.Literal[PR.GET_NAME], arg2: Pointer[str]) -> ReadablePointer[str]: ...
    @t.overload
    async def prctl(self, option: t.Literal[PR.SET_NAME], arg2: WrittenPointer[str]) -> None: ...

    async def prctl(self, option: PR, arg2, arg3=0, arg4=0, arg5=0) -> t.Union[int, Pointer, None]:
        """operations on a process or thread

        This has overloads for each prctl option, so it's type-safe to use this method.

        man: prctl(2)
        """
        if option is PR.SET_PDEATHSIG:
            return await self.prctl_set_pdeathsig(option, arg2)
        elif option is PR.CAP_AMBIENT:
            return await self.prctl_cap_ambient(option, arg2, arg3)
        elif option is PR.GET_NAME:
            return await self.prctl_get_name(option, arg2)
        elif option is PR.SET_NAME:
            return await self.prctl_set_name(option, arg2)
        else:
            return await _prctl(self.sysif, option, arg2, arg3, arg4, arg5)

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _prctl(sysif: SyscallInterface, option: PR, arg2: t.Union[int, near.Address],
                 arg3: t.Optional[t.Union[int, near.Address]]=None, arg4: t.Optional[t.Union[int, near.Address]]=None,
                 arg5: t.Optional[t.Union[int, near.Address]]=None) -> int:
    if arg3 is None:
        arg3 = 0
    if arg4 is None:
        arg4 = 0
    if arg5 is None:
        arg5 = 0
    return (await sysif.syscall(SYS.prctl, option, arg2, arg3, arg4, arg5))



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.sys.timerfd import *

class TestTimerfd(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread
        self.fd = await self.thr.task.timerfd_create(CLOCK.REALTIME)

    async def asyncTearDown(self) -> None:
        await self.fd.close()

    async def test(self) -> None:
        await self.fd.timerfd_settime(
            TFD_TIMER.NONE, await self.thr.ram.ptr(Itimerspec(Timespec(0, 0), Timespec(0, 1))))
        await self.fd.timerfd_gettime(await self.thr.ram.malloc(Itimerspec))



"""Definitions of namespace-local identifiers.

These namespace-local identifiers are like near pointers, in systems
with segmented memory. They are valid only within a specific segment
(namespace).

"""

from __future__ import annotations
from dataclasses import dataclass

__all__  = [
    "FileDescriptor",
    "WatchDescriptor",
    "Address",
    "MemoryMapping",
    "Process",
    "ProcessGroup",
]

@dataclass(frozen=True)
class FileDescriptor:
    """The integer identifier for a file descriptor taken by many syscalls.

    This is a file descriptor in a specific file descriptor table, but we don't with this
    object know what file descriptor table that is.

    """
    __slots__ = ('number')
    number: int

    def __str__(self) -> str:
        return f"FD({self.number})"

    def __repr__(self) -> str:
        return str(self)

    def __int__(self) -> int:
        return self.number

@dataclass(frozen=True)
class WatchDescriptor:
    """The integer identifier for an inotify watch descriptor taken by inotify syscalls.

    This is a watch descriptor for a specific inotify instance, but we don't with this
    object know what inotify instance that is.

    """
    number: int

    def __str__(self) -> str:
        return f"WD({self.number})"

    def __repr__(self) -> str:
        return str(self)

    def __int__(self) -> int:
        return self.number

@dataclass(frozen=True)
class Address:
    """The integer identifier for a virtual memory address as taken by many syscalls.

    This is an address in a specific address space, but we don't with this object know
    what address space that is.

    """
    __slots__ = ('address')
    address: int

    def __add__(self, other: int) -> 'Address':
        return Address(self.address + other)

    def __sub__(self, other: int) -> 'Address':
        return Address(self.address - other)

    def __str__(self) -> str:
        return f"Address({hex(self.address)})"

    def __repr__(self) -> str:
        return str(self)

    def __int__(self) -> int:
        return self.address

@dataclass(frozen=True)
class MemoryMapping:
    """The integer identifiers for a virtual memory mapping as taken by many syscalls.

    This is a mapping in a specific address space, but we don't with this object know what
    address space that is.

    We require three pieces of information to describe a memory mapping. 
    - Address is the start address of the memory mapping
    - Length is the length in bytes of the memory mapped region

    Page size is unusual, but required for robustness: While the syscalls related to
    memory mappings don't appear to depend on page size, that's an illusion. They seem to
    deal in sizes in terms of bytes, but if you provide a size which is not a multiple of
    the page size, silent failures or misbehaviors will occur. Misbehavior include the
    sizes being rounded up to the page size, including in munmap, thus unmapping more
    memory than expected.

    As long as we ensure that the original length we pass to mmap is a multiple of the
    page size that will be used for the mapping, then we could get by with just storing
    the length and not the page size. However, the memory mapping API allows unmapping
    only part of a mapping, or in general performing operations on only part of a
    mapping. These splits must happen at page boundaries, and therefore to support
    specifying these splits without allowing silent rounding errors, we need to know the
    page size of the mapping.

    This is especially troubling when mmaping files with an unknown page size, such as
    those passed to us from another program. memfd_create or hugetlbfs can be used to
    create files with an unknown page size, which cannot be robust unmapped. At this time,
    we don't know of a way to learn the page size of such a file. One good solution would
    be for mmap to be taught a new MAP_ENFORCE_PAGE_SIZE flag which requires MAP_HUGE_* to
    be passed when mapping files with nonstandard page size. In this way, we could assert
    the page size of the file and protect against attackers sending us files with
    unexpected page sizes.

    """
    __slots__ = ('address', 'length', 'page_size')
    address: int
    length: int
    page_size: int

    def __post_init_(self) -> None:
        if (self.address % self.page_size) != 0:
            raise Exception("the address for this memory-mapping is not page-aligned", self)
        if (self.length % self.page_size) != 0:
            raise Exception("the length for this memory-mapping is not page-aligned", self)

    def as_address(self) -> Address:
        "Return the starting address of this memory mapping."
        return Address(self.address)

    def __str__(self) -> str:
        if self.page_size == 4096:
            return f"MMap({hex(self.address)}, {self.length})"
        else:
            return f"MMap(pgsz={self.page_size}, {hex(self.address)}, {self.length})"

    def __repr__(self) -> str:
        return str(self)

@dataclass(frozen=True)
class Process:
    """The integer identifier for a process taken by many syscalls.

    This is a process in a specific pid namespace, but we don't with this object know what
    pid namespace that is.

    """
    id: int

    def __int__(self) -> int:
        return self.id

    def __str__(self) -> str:
        return f'Process({self.id})'

    def __repr__(self) -> str:
        return str(self)

@dataclass(frozen=True)
class ProcessGroup:
    """The integer identifier for a process group taken by many syscalls.

    This is a process group in a specific pid namespace, but we don't with this object
    know what pid namespace that is.

    """
    id: int

    def __int__(self) -> int:
        return self.id



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread

from rsyscall.sys.inotify import *
from rsyscall.inotify_watch import Inotify
from rsyscall.fcntl import O
from rsyscall.stdlib import mkdtemp

class TestInotify(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread
        self.tmpdir = await mkdtemp(self.thr)
        self.ify = await Inotify.make(self.thr)

    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def test_create(self) -> None:
        watch = await self.ify.add(self.tmpdir, IN.CREATE)
        name = "foo"
        fd = await self.thr.task.open(await self.thr.ram.ptr(self.tmpdir/name), O.CREAT|O.EXCL)
        event = await watch.wait_until_event(IN.CREATE, name)
        self.assertEqual(event.name, name)
        self.assertEqual(event.mask, IN.CREATE)



"Provides the Command class, which is a convenient representation of the arguments to execve."
import typing as t
from rsyscall.path import Path
import os

T_command = t.TypeVar('T_command', bound="Command")
class Command:
    "A convenient builder-pattern representation of the arguments to execve."
    def __init__(self,
                 executable_path: t.Union[str, os.PathLike],
                 arguments: t.List[t.Union[str, os.PathLike]],
                 env_updates: t.Mapping[str, t.Union[str, os.PathLike]]) -> None:
        self.executable_path = executable_path
        self.arguments = arguments
        self.env_updates = env_updates

    def args(self: T_command, *args: t.Union[str, os.PathLike]) -> T_command:
        "Add more arguments to this Command."
        return type(self)(self.executable_path,
                          [*self.arguments, *args],
                          self.env_updates)

    def env(self: T_command, env_updates: t.Mapping[str, t.Union[str, os.PathLike]]={},
            **updates: t.Union[str, os.PathLike]) -> T_command:
        """Add more environment variable updates to this Command.

        There are two ways to pass arguments to this method (which can be used simultaneously):
        - you can pass a dictionary of environment updates,
        - or you can provide your environment updates as keyword arguments.
        Both are necessary, since there are many valid environment variable
        names which are not valid Python keyword argument names.

        """
        return type(self)(self.executable_path,
                          self.arguments,
                          {**self.env_updates, **env_updates, **updates})

    def in_shell_form(self) -> str:
        "Render this Command as a string which could be passed to a shell."
        ret = ""
        for key, value in self.env_updates.items():
            ret += os.fsdecode(key) + "=" + os.fsdecode(value) + " "
        ret += os.fsdecode(self.executable_path)
        # skip first argument
        for arg in self.arguments[1:]:
            ret += " " + os.fsdecode(arg)
        return ret

    def __str__(self) -> str:
        ret = "Command("
        for key, value in self.env_updates.items():
            ret += f"{key}={value} "
        ret += f"{os.fsdecode(self.executable_path)},"
        for arg in self.arguments:
            ret += " " + os.fsdecode(arg)
        ret += ")"
        return ret

    def __repr__(self) -> str:
        return str(self)



"""Definitions of namespace-local identifiers, syscalls, and SyscallInterface

These namespace-local identifiers are like near pointers, in systems
with segmented memory. They are valid only within a specific segment
(namespace).

The syscalls are instructions, operating on near pointers and other
arguments.

The SyscallInterface is the segment register override prefix, which is
used with the instruction to say which segment register to use for the
syscall.

We don't know from a segment register override prefix alone that the
near pointers we are passing to an instruction are valid pointers in
the segment currently contained in the segment register.

In terms of our actual classes: We don't know from a SyscallInterface
alone that the identifiers we are passing to a syscall match the
namespaces active in the task behind the SyscallInterface.

(The task is like the segment register, in this analogy.)

"""
# re-exported namepsace-local identifiers
from rsyscall.near.types import (
    FileDescriptor,
    WatchDescriptor,
    Address,
    MemoryMapping,
    Process,
    ProcessGroup,
)
# re-exported SyscallInterface
from rsyscall.near.sysif import SyscallInterface, SyscallHangup
__all__ = [
    'FileDescriptor',
    'WatchDescriptor',
    'Address',
    'MemoryMapping',
    'Process',
    'ProcessGroup',
    'SyscallInterface', 'SyscallHangup',
]



"Defines AllocationInterface."
from __future__ import annotations
import abc
import typing as t

class UseAfterFreeError(Exception):
    pass

class AllocationInterface:
    """Represents an allocation of a range of bytes in some file

    The file is typically mapped into memory. Essentially, this is a single allocation
    returned by malloc; but it's not necessarily tied to memory.

    """
    @abc.abstractmethod
    def offset(self) -> int:
        """Get the offset of this allocation in its memory mapping; throws if this allocation has been invalidated.

        TODO this should return the offset of the allocation

        """
        pass
    @abc.abstractmethod
    def size(self) -> int:
        "Get the size of this allocation."
        pass
    @abc.abstractmethod
    def split(self, size: int) -> t.Tuple[AllocationInterface, AllocationInterface]:
        """Invalidate this allocation and split it into two adjacent allocations.

        These two allocations can be independently freed, or split again, ad infinitum;
        they can also be merged back together with merge.

        """
        pass
    @abc.abstractmethod
    def merge(self, other: AllocationInterface) -> AllocationInterface:
        """Invalidate these two adjacent allocations and merge them into one; only works if they came from split.

        Call this on the left allocation returned from split, and pass the right allocation.

        Depending on the characteristics of the underlying allocator, this may also work
        for two unrelated allocations rather than just ones that came from split, but you
        certainly shouldn't try.

        """
        pass
    @abc.abstractmethod
    def free(self) -> None:
        "Invalidate this allocation and return its range for re-allocation; also called automatically on __del__."
        pass




"A program developed for SIGBOVIK2020; see research/sigbovik2020 for more information."
from __future__ import annotations
import shlex

from rsyscall.sched import CLONE
from rsyscall.fcntl import O
from rsyscall.unistd import SEEK
from rsyscall.path import Path
from rsyscall.stdlib import mkdtemp
from rsyscall.sys.mount import MS, UMOUNT
from rsyscall.sys.stat import Stat
from rsyscall.near.sysif import SyscallError

from rsyscall.linux.fuse import (
    FuseOut,
    FuseInitOp, FuseLookupOp, FuseOpenOp, FuseOpendirOp, FuseReadOp, FuseGetattrOp, FuseReaddirplusOp, FuseFlushOp, FuseReleaseOp, FuseReleasedirOp, FuseReadlinkOp, FuseGetxattrOp,
    FuseInList, FuseInitIn, FuseInitOut, FuseAttr, FuseEntryOut, FuseOpenOut, FOPEN,
    FuseAttrOut,
    FuseDirentplus, FuseDirent,
    FUSE_INIT,
    FUSE_MIN_READ_BUFFER,
)
from rsyscall.linux.dirent import DirentList, DT
from rsyscall.time import Timespec
from rsyscall.sys.stat import TypeMode, S_IF, Mode
from rsyscall import Command, WrittenPointer, FileDescriptor, Thread
from rsyscall.memory.ram import RAM
from rsyscall.unistd import AT, ArgList
from rsyscall.tasks.stub import StubServer
import errno
import os
import trio
import typing as t
import logging

logger = logging.getLogger(__name__)

from dataclasses import dataclass
@dataclass
class CommandData:
    args: t.List[str]
    stdin: Stat
    stdout: Stat
    stderr: Stat

def reconstruct(commands: t.List[CommandData]) -> str:
    script = ""
    for i in range(len(commands)):
        if i == 0:
            # there's no guarantee that they'll exec in order, lol, but this makes for simpler code
            script += " ".join(commands[0].args)
            stdin = commands[0].stdin
        else:
            prev = commands[i-1]
            cur = commands[i]
            if prev.stdout.ino == cur.stdin.ino and cur.stdin.ino != stdin.ino:
                # it also doubles as a pretty printer
                script += " | "
            else:
                script += "\n"
            script += " ".join(commands[i].args)
    script += "\n"
    return script

import argparse

class FuseFS:
    @classmethod
    async def mount(cls, thread: Thread, path: Path) -> FuseFS:
        self = cls()
        await self._mount(thread, path)
        return self

    async def _mount(self, thread: Thread, path: Path) -> None:
        # TODO does a pid namespace kill a thread in D-wait?
        # TODO we could make a better API here with the new mount API
        self.path = path
        # /dev/fuse doesn't seem to behave nicely when used with EPOLLET, so we'll just do blocking
        # IO on it in a dedicated thread - we need the dedicated thread anyway (see comment below
        # about private fd table) so it's not extra overhead.
        devfuse = await thread.task.open(await thread.ptr(Path("/dev/fuse")), O.RDWR)
        self.uid = await thread.task.getuid()
        self.gid = await thread.task.getgid()
        self.parent_thread = thread
        await thread.mount("ignored", self.path, "fuse", MS.NONE,
                           f"fd={int(devfuse)},rootmode=40777,user_id={self.uid},group_id={self.gid}")
        # We'll keep devfuse open *only* in the dedicated server thread's private fd table, so that
        # other threads accessing the filesystem don't deadlock when we abort the FUSE server loop -
        # instead their syscalls will be aborted with ENOTCONN.
        self.thr = await thread.clone()
        self.devfuse = self.thr.task.inherit_fd(devfuse)
        await devfuse.close()
        # Respond to FUSE init message to sanity-check things are set up right.
        self.buf = await self.thr.malloc(FuseInList, FUSE_MIN_READ_BUFFER)
        [init] = await self.read()
        if not isinstance(init, FuseInitOp):
            raise Exception("oops, got non-init as first message", init)
        flags = init.msg.flags & ~(FUSE_INIT.EXPORT_SUPPORT)
        # /dev/fuse doesn't ever do partial writes, nice.
        await self.write(init.respond(FuseInitOut(
            major=init.msg.major, minor=init.msg.minor, max_readahead=init.msg.max_readahead,
            flags=flags,
            max_background=16, congestion_threshold=16, max_write=128, time_gran=128)))
        # Note that deadlocks are still possible, if the root thread makes blocking calls into the
        # filesystem. As usual, we should avoid doing work on the root thread, and instead do it in
        # children.

    async def read(self) -> FuseInList:
        # /dev/fuse only returns complete packets, so we don't need to rebuffer, wonderful.
        read, unused = await self.devfuse.read(self.buf)
        msgs = await read.read()
        logger.debug("read in FUSE messages: %s", msgs)
        self.buf = read + unused
        return msgs

    async def write(self, out: FuseOut) -> None:
        logger.debug("writing out FUSE response: %s", out)
        ptr = await self.thr.ptr(out)
        written, unwritten = await self.devfuse.write(ptr)
        if unwritten.size() != 0:
            raise Exception("/dev/fuse is not supposed to ever do partial writes, but I got one somehow on", ptr)

    async def __aenter__(self) -> FuseFS:
        return self

    async def cleanup(self) -> None:
        # umount to kill the fuse loop; it's a bit lazy to save the parent thread to do this, but we
        # can't do it in self.thr because that thread spends all its time blocked in read.
        await self.parent_thread.task.umount(await self.parent_thread.ptr(self.path))
        # and exit the thread - but it might already be dead, so we might fail to exit it, just ignore that.
        try:
            await self.thr.exit(0)
        except SyscallError:
            pass

    async def __aexit__(self, *args, **kwargs) -> None:
        await self.cleanup()

    def __init__(self) -> None:
        pass

async def symsh_main(thread: Thread, command: Command) -> None:
    async with (await mkdtemp(thread)) as tmpdir:
        thr = await thread.clone(flags=CLONE.NEWUSER|CLONE.NEWNS)
        path = tmpdir/"path"
        await thr.mkdir(path)
        async with trio.open_nursery() as nursery:
            stub_server = await StubServer.make(thr, tmpdir, "stub")
            stub_path = tmpdir/"stub"
            commands: t.List[CommandData] = []
            @nursery.start_soon
            async def exec_server() -> None:
                """Accepts connections from the stub executables

                We just check the command and stat stdin/stdout/stderr, then exit the stub.

                """
                while True:
                    args, thread = await stub_server.accept()
                    buf = await thread.malloc(Stat)
                    async def stat(fd: FileDescriptor) -> Stat:
                        nonlocal buf
                        buf = await fd.fstat(buf)
                        return await buf.read()
                    name, *rest = args
                    commands.append(CommandData(
                        args=[Path(name).name, *rest],
                        stdin=await stat(thread.stdin),
                        stdout=await stat(thread.stdout),
                        stderr=await stat(thread.stderr),
                    ))
                    await thread.exit(0)
            async with (await FuseFS.mount(thr, path)) as fuse:
                @nursery.start_soon
                async def open() -> None:
                    "Every filename in this filesystem is a symlink to stub_path"
                    while True:
                        try:
                            [op] = await fuse.read()
                        except OSError as e:
                            # sure wish Python had a builtin ENODEV exception, that would come in handy
                            if e.errno == errno.ENODEV:
                                # the filesystem has been unmounted, just return cleanly
                                return
                            else:
                                raise
                        sym_ino = 2
                        fh = 42
                        if isinstance(op, FuseGetattrOp):
                            await fuse.write(op.respond(FuseAttrOut(
                                attr_valid=Timespec(10000, 0),
                                attr=FuseAttr(
                                    ino=1, size=0, blocks=1,
                                    atime=Timespec(0, 0), mtime=Timespec(0, 0), ctime=Timespec(0, 0),
                                    mode=TypeMode(S_IF.DIR, Mode(0o777)), nlink=1,
                                    uid=fuse.uid, gid=fuse.gid, rdev=0, blksize=4096
                                ))))
                        elif isinstance(op, FuseLookupOp):
                            await fuse.write(op.respond(FuseEntryOut(
                                nodeid=sym_ino, generation=1,
                                entry_valid=Timespec(10000, 0), attr_valid=Timespec(10000, 0),
                                # the size needs to be consistent with the data we'll actually send back on read
                                # the kernel, I guess, handles delivering an eof
                                attr=FuseAttr(
                                    ino=999, size=16, blocks=1,
                                    atime=Timespec(0, 0), mtime=Timespec(0, 0), ctime=Timespec(0, 0),
                                    mode=TypeMode(S_IF.LNK, Mode(0o777)), nlink=1,
                                    uid=fuse.uid, gid=fuse.gid, rdev=0, blksize=4096
                                ))))
                        elif isinstance(op, FuseOpenOp):
                            await fuse.write(op.respond(FuseOpenOut(fh=fh, open_flags=FOPEN.NONE)))
                        elif isinstance(op, FuseGetxattrOp):
                            await fuse.write(op.error(-errno.ENODATA))
                        elif isinstance(op, FuseReadOp):
                            await fuse.write(op.respond(bytes(16)))
                        elif isinstance(op, FuseReleaseOp):
                            await fuse.write(op.respond())
                        elif isinstance(op, FuseReadlinkOp):
                            await fuse.write(op.respond(stub_path))
                        else:
                            print("unhandled op", op)
                            raise Exception("unhandled op", op)
                # now we simply run the command with PATH set to only our FUSE path, and boom
                await thr.run(command.env(PATH=path))
                print("==================== reconstructed script ====================")
                print(reconstruct(commands), end='')
            nursery.cancel_scope.cancel()

async def amain(argv: t.List[str]) -> None:
    # hmm I should write a more TFS/direct style argument parser for python
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='subcommand', required=True)
    
    example_parser = subparsers.add_parser('example', help='Example')
    exec_parser = subparsers.add_parser('exec', help="Exec the provided executable, no arguments allowed")
    exec_parser.add_argument(dest='executable')
    
    args = parser.parse_args(argv[1:])
    from rsyscall import local_thread
    if args.subcommand == 'example':
        script = """ls; which ls
stat /
foo|bar
"""
        print("======================= running script =======================")
        print(script, end='')
        await symsh_main(local_thread, local_thread.environ.sh.args("-c", script))
    elif args.subcommand == 'exec':
        await symsh_main(local_thread, Command(Path(args.executable), [args.executable], {}))

def main():
    import sys
    trio.run(amain, sys.argv)

if __name__ == "__main__":
    main()



import typing as t
import trio

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface, SyscallHangup
from rsyscall.sys.syscall import SYS

async def _execve(sysif: SyscallInterface,
                  path: near.Address, argv: near.Address, envp: near.Address) -> None:
    def handle(exn):
        if isinstance(exn, SyscallHangup):
            return None
        else:
            return exn
    with trio.MultiError.catch(handle):
        await sysif.syscall(SYS.execve, path, argv, envp)

async def _execveat(sysif: SyscallInterface,
                    dirfd: t.Optional[near.FileDescriptor], path: near.Address,
                    argv: near.Address, envp: near.Address, flags: int) -> None:
    if dirfd is None:
        dirfd = AT.FDCWD # type: ignore
    def handle(exn):
        if isinstance(exn, SyscallHangup):
            return None
        else:
            return exn
    with trio.MultiError.catch(handle):
        await sysif.syscall(SYS.execveat, dirfd, path, argv, envp, flags)

async def _exit(sysif: SyscallInterface, status: int) -> None:
    def handle(exn):
        if isinstance(exn, SyscallHangup):
            return None
        else:
            return exn
    with trio.MultiError.catch(handle):
        await sysif.syscall(SYS.exit, status)



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread

from rsyscall.nix import *
from rsyscall.sched import CLONE
from rsyscall.stdlib import mkdtemp
import rsyscall._nixdeps.nix

class TestNix(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.tmpdir = await mkdtemp(local_thread)
        self.thr = await local_thread.clone()
        await enter_nix_container(local_thread, rsyscall._nixdeps.nix.closure, self.thr, self.tmpdir)

    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def test_hello(self) -> None:
        hello = (await deploy(self.thr, rsyscall._nixdeps.coreutils.closure)).bin('echo').args('hello world')
        await self.thr.run(hello)

    async def test_with_daemon(self) -> None:
        nix_daemon = (await deploy(self.thr, rsyscall._nixdeps.nix.closure)).bin("nix-daemon")
        nd_child = await (await self.thr.clone()).exec(nix_daemon)
        self.thr.environ['NIX_REMOTE'] = 'daemon'
        hello = (await deploy(self.thr, rsyscall._nixdeps.coreutils.closure)).bin('echo').args('hello world')
        await self.thr.run(hello)



from __future__ import annotations
import ipaddress
import time
import os
import abc
import trio
import rsyscall.handle as handle
import rsyscall.near
from rsyscall.trio_test_case import TrioTestCase
from rsyscall.thread import Thread
from rsyscall.command import Command
from rsyscall.handle import FileDescriptor, Path
from dataclasses import dataclass
import dns.zone
from dns.rdataset import from_text_list as make_rdset
import rsysapps.dnspython_LUA
import typing as t
dns.rdata.register_type(rsysapps.dnspython_LUA, 65402, 'LUA')

from rsyscall.netinet.in_ import SockaddrIn
from rsyscall.sys.socket import AF, SOCK
from rsyscall.sched import CLONE

@dataclass
class Powerdns:
    pass

async def start_powerdns(nursery, parent: Thread, path: Path, zone: dns.zone.Zone,
                                  # tuple is (udpfd, listening tcpfd)
                                  ipv4_sockets: t.List[t.Tuple[handle.FileDescriptor, handle.FileDescriptor]],
                                  ipv6_sockets: t.List[t.Tuple[handle.FileDescriptor, handle.FileDescriptor]],
) -> Powerdns:
    pdns_server = Command(Path("/home/sbaugh/.local/src/pdns/pdns/pdns_server"), ['pdns_server'], {})
    # pdns_server = await parent.environ.which("pdns_server")
    thread = await parent.clone()

    # we pretend to pass addresses like 0.0.0.1 etc
    # we add one so we don't pass 0.0.0.0 and make powerdns think it's bound to everything
    ipv4s = {str(i+1): (udp.move(thread.task), tcp.move(thread.task))
             for i, (udp, tcp) in enumerate(ipv4_sockets)}
    ipv6s = {str(i+1): (udp.move(thread.task), tcp.move(thread.task))
             for i, (udp, tcp) in enumerate(ipv6_sockets)}
    await thread.unshare_files(going_to_exec=True)
    config = {
        "config-dir": os.fsdecode(path),
        # TODO control-console seems to be a feature where it will listen on stdin or something?
        # we should use that instead of this socketdir
        "socket-dir": os.fsdecode(path),
        # more logging
        "loglevel": "9",
        "log-dns-queries": "yes",
        # backend
        "launch": "bind",
        "bind-config": os.fsdecode(await thread.spit(path/"named.conf",
            'zone "%s" { file "%s"; };' % (
                zone.origin.to_text(),
                os.fsdecode(await thread.spit(path/"zone", zone.to_text()))))),
        "enable-lua-records": "yes",
        # relevant stuff
        "local-address": ",".join(ipv4s),
        "local-address-udp-fds": ",".join([f"{i}={await fd.as_argument()}" for i, (fd, _) in ipv4s.items()]),
        "local-address-tcp-fds": ",".join([f"{i}={await fd.as_argument()}" for i, (_, fd) in ipv4s.items()]),
        "local-ipv6": ",".join(ipv6s),
        "local-ipv6-udp-fds": ",".join([f"{i}={await fd.as_argument()}" for i, (fd, _) in ipv6s.items()]),
        "local-ipv6-tcp-fds": ",".join([f"{i}={await fd.as_argument()}" for i, (_, fd) in ipv6s.items()]),
    }
    print(config['local-address-udp-fds'])
    child = await thread.exec(pdns_server.args(*[f"--{name}={value}" for name, value in config.items()]))
    nursery.start_soon(child.check)
    return Powerdns()

async def start_recursor(nursery, parent: Thread, path: Path,
                                  ipv4_sockets: t.List[t.Tuple[handle.FileDescriptor, handle.FileDescriptor]],
                                  ipv6_sockets: t.List[t.Tuple[handle.FileDescriptor, handle.FileDescriptor]],
                                  root_hints: dns.zone.Zone=None) -> Powerdns:
    pdns_recursor = Command(Path("/home/sbaugh/.local/src/pdns/pdns/recursordist/pdns_recursor"), ['pdns_recursor'], {})
    thread = await parent.clone()

    ipv4s = {str(i+1): (udp.move(thread.task), tcp.move(thread.task))
             for i, (udp, tcp) in enumerate(ipv4_sockets)}
    ipv6s = {"::"+str(i+1): (udp.move(thread.task), tcp.move(thread.task))
             for i, (udp, tcp) in enumerate(ipv6_sockets)}
    addresses = {**ipv4s, **ipv6s}
    await thread.unshare_files(going_to_exec=True)
    config = {
        "config-dir": os.fsdecode(path),
        "socket-dir": os.fsdecode(path),
        # more logging
        "loglevel": "9",
        "log-common-errors": "yes",
        "quiet": "no",
        "trace": "yes",
        "dont-query": "",
        "logging-facility": "0",
        # relevant stuff
        "local-address": ",".join(addresses),
        "allow-from": "127.0.0.0/8",
        "local-address-udp-fds": ",".join([f"{i}={await fd.as_argument()}" for i, (fd, _) in addresses.items()]),
        "local-address-tcp-fds": ",".join([f"{i}={await fd.as_argument()}" for i, (_, fd) in addresses.items()]),
    }
    if root_hints is not None:
        config["hint-file"] = os.fsdecode(await thread.spit(path/'root.hints', root_hints.to_text()))
    child = await thread.exec(pdns_recursor.args(*[f"--{name}={value}" for name, value in config.items()]))
    nursery.start_soon(child.check)
    return Powerdns()

def make_node(rdatasets: t.List[dns.rdataset.Rdataset]) -> dns.node.Node:
    node = dns.node.Node()
    node.rdatasets = rdatasets
    return node

def make_zone(origin: t.Union[str, dns.name.Name],
              nodes: t.Dict[t.Union[str, dns.name.Name], dns.rdataset.Rdataset]) -> dns.zone.Zone:
    zone = dns.zone.Zone(origin)
    for name, rdatasets in nodes.items():
        zone[name] = make_node(rdatasets)
    return zone


import rsyscall.tasks.local as local
class TestPowerdns(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thread = await local.thread.clone()
        await self.thread.unshare_user(0, 0)
        await self.thread.unshare(CLONE.NEWNET)
        # set loopback up
        ip = await self.thread.environ.which("ip")
        # TODO blah this requires root and we don't want to have to run as root because it's a hassle
        # well, I guess we should be sandboxing anything bad, so maybe we should run as root..
        # no that's silly
        await self.thread.run(ip.args('link', 'set', 'dev', 'lo', 'up'))
        self.pdns_idx = 1
        self.tmpdir = await self.thread.mkdtemp("test_powerdns")
        self.path = self.tmpdir.path
        self.dig = await self.thread.environ.which("dig")
        self.host = await self.thread.environ.which("host")

    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def start_authoritative(self, zone: dns.zone.Zone) -> ipaddress.IPv4Address:
        idx = self.pdns_idx
        self.pdns_idx += 1
        addr = ipaddress.IPv4Address(0x7F_00_00_00 + idx)
        sockaddr = await self.thread.ram.ptr(SockaddrIn(53, addr))
        print("sockaddr", sockaddr)
        path = await self.thread.mkdir(self.path/f"pdns{idx}")

        udp_sock = await self.thread.task.socket(AF.INET, SOCK.DGRAM)
        await udp_sock.bind(sockaddr)
        tcp_sock = await self.thread.task.socket(AF.INET, SOCK.STREAM)
        await tcp_sock.bind(sockaddr)
        await tcp_sock.listen(10)

        def add_rdata(nodename: str, typ: str, ttl: int, data: str) -> None:
            node = zone.find_node(nodename, create=True)
            typ_int = dns.rdatatype.from_text(typ)
            rdata = dns.rdata.from_text(dns.rdataclass.IN, typ_int, data)
            node.find_rdataset(dns.rdataclass.IN, typ_int, create=True).add(rdata, ttl=ttl)
        add_rdata('@', 'SOA', 3600, 'ns1 root 1 123 123 123 123')
        add_rdata('@', 'NS', 3600, 'ns1')
        add_rdata('ns1', 'A', 3600, str(addr))
        powerdns = await start_powerdns(
            self.nursery, self.thread, path, zone, [(udp_sock, tcp_sock)], [])
        return addr

    async def start_recursor(self, root_hints: dns.zone.Zone) -> ipaddress.IPv4Address:
        idx = self.pdns_idx
        self.pdns_idx += 1
        addr = ipaddress.IPv4Address(0x7F_00_00_00 + idx)
        sockaddr = await self.thread.ram.ptr(SockaddrIn(53, addr))
        path = await self.thread.mkdir(self.path/f"pdns{idx}")

        udp_sock = await self.thread.task.socket(AF.INET, SOCK.DGRAM)
        await udp_sock.bind(sockaddr)
        tcp_sock = await self.thread.task.socket(AF.INET, SOCK.STREAM)
        await tcp_sock.bind(sockaddr)
        await tcp_sock.listen(10)

        powerdns = await start_recursor(
            self.nursery, self.thread, path, [(udp_sock, tcp_sock)], [], root_hints=root_hints)
        return addr

    async def test_basic(self) -> None:
        user_addr = await self.start_authoritative(make_zone('user', {
            'a': [make_rdset('IN', 'A', 3600, ['1.3.5.7'])],
        }))
        root_addr = await self.start_authoritative(make_zone('.', {
            'user': [make_rdset('IN', 'NS', 3600, ['user.glue.neato.'])],
            'user.glue.neato': [make_rdset('IN', 'A', 3600, [str(user_addr)])],
        }))
        recursor_addr = await self.start_recursor(make_zone('.', {
            '.': [make_rdset('IN', 'NS', 3600000, ['A.ROOT-SERVERS.NET.'])],
            'a.root-servers.net': [make_rdset('IN', 'A', 3600000, [str(root_addr)])],
        }))

        # TODO should assert that the address returned is actually correct
        await self.thread.run(self.host.args('a.user', str(recursor_addr)))

    async def test_bad_ns(self) -> None:
        magic_addr = await self.start_authoritative(make_zone('.', {
            '*': [make_rdset('IN', 'NS', 3600, ['foo.bar.'])],
            # 'magical': [make_rdset('IN', 'NS', 3600, ['foo.bar.'])],
            # '*': [make_rdset('IN', 'LUA', 3600, ["NS \"; return 'foo.bar.';\""])],
        }))
        await self.thread.run(self.dig.args('@'+str(magic_addr), 'NS', 'magical'))

    async def test_powerdns(self) -> None:
        user_addr = await self.start_authoritative(make_zone('user.1.magic', {
            'a': [make_rdset('IN', 'A', 3600, ['1.3.5.7'])],
        }))
        magic_addr = await self.start_authoritative(make_zone('magic', {
            # 'a.user.1': [make_rdset('IN', 'NS', 3600, ['user.'])]
            '*': [make_rdset('IN', 'LUA', 3600, ["""NS (
"; local dn = qname:makeRelative(newDN('magic'));"
"local labels = dn:getRawLabels();"
"local components = labels[#labels];"
"if tonumber(components) > (#labels - 1) then return 'no.domain' end;"
"local nslabels = {table.unpack(labels, #labels-components, #labels-1)};"
"return table.concat(nslabels, '.');"
)"""])]
        }))
        real_user_addr = await self.start_authoritative(make_zone('user', {
            '@': [make_rdset('IN', 'A', 3600, [str(user_addr)])],
        }))
        root_addr = await self.start_authoritative(make_zone('.', {
            'magic': [make_rdset('IN', 'NS', 3600, ['magic.glue.neato.'])],
            'magic.glue.neato': [make_rdset('IN', 'A', 3600, [str(magic_addr)])],
            'user': [make_rdset('IN', 'NS', 3600, ['user.glue.neato.'])],
            'user.glue.neato': [make_rdset('IN', 'A', 3600, [str(real_user_addr)])],
        }))
        recursor_addr = await self.start_recursor(make_zone('.', {
            '.': [make_rdset('IN', 'NS', 3600000, ['A.ROOT-SERVERS.NET.'])],
            'a.root-servers.net': [make_rdset('IN', 'A', 3600000, [str(root_addr)])],
        }))

        self.host = await self.thread.environ.which("host")
        # oh hmm we want to return an NS record for something else though
        # that's awkward to do with lua records.
        # or... maybe this is fine? can we return NS record for this domain at all levels?
        # in any case, now we have this thing working.
        # so... we can write a better test.
        # await self.thread.run(dig.args('@127.0.0.2', 'NS', 'a.root-servers.net'))
        # await self.thread.run(self.dig.args('@'+str(user_addr), 'A', 'a.user.1.magic'))
        # await self.thread.run(self.dig.args('@'+str(recursor_addr), '+trace', 'NS', 'user'))
        # await self.thread.run(self.dig.args('@'+str(recursor_addr), 'A', 'a.user.1.magic'))
        await self.thread.run(self.dig.args('@'+str(magic_addr), 'NS', 'a.user.1.magic'))
        # await self.thread.run(self.dig.args('@'+str(recursor_addr), 'A', 'user'))
        # await self.thread.run(self.dig.args('@'+str(recursor_addr), 'A', 'a.user.1.magic'))
        # await self.thread.run(self.host.args('-v', '-a', 'a.user.1.magic', str(recursor_addr)))
        # await self.thread.run(self.host.args('-a', 'a.user.1.magic', str(magic_addr)))
        # raise Exception('hi')



from __future__ import annotations
from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.tasks.persistent import *
from rsyscall.tasks.ssh import make_local_ssh
from rsyscall.near.sysif import SyscallHangup
from rsyscall.tests.utils import assert_thread_works
from rsyscall.sched import CLONE
from rsyscall.signal import SIG
from rsyscall.stdlib import mkdtemp
import unittest

class TestPersistent(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thread = local_thread
        self.tmpdir = await mkdtemp(self.thread, "test_stub")
        self.sock_path = self.tmpdir/"persist.sock"

    async def asyncTearDown(self) -> None:
        await self.tmpdir.cleanup()

    async def test_reconnect_exit(self) -> None:
        per_thr = await clone_persistent(self.thread, self.sock_path)
        await assert_thread_works(self, per_thr)
        await per_thr.reconnect(self.thread)
        await assert_thread_works(self, per_thr)
        await per_thr.exit(0)

    async def test_exit_reconnect(self) -> None:
        thread = await self.thread.clone()
        per_thr = await clone_persistent(self.thread, self.sock_path)
        await per_thr.prep_for_reconnect()
        await per_thr.exit(0)
        # when we try to reconnect, we'll fail
        with self.assertRaises(ConnectionRefusedError):
            await per_thr.reconnect(self.thread)

    async def test_nest_exit(self) -> None:
        per_thr = await clone_persistent(self.thread, self.sock_path)
        thread = await per_thr.clone(CLONE.FILES)
        await per_thr.reconnect(self.thread)
        await assert_thread_works(self, thread)
        await thread.exit(0)

    async def test_nest_unshare_files_exit(self) -> None:
        per_thr = await clone_persistent(self.thread, self.sock_path)
        thread = await per_thr.clone()
        await per_thr.reconnect(self.thread)
        await assert_thread_works(self, thread)
        await thread.exit(0)

    async def test_ssh_same(self) -> None:
        host = await make_local_ssh(self.thread)
        local_child, remote_thread = await host.ssh(self.thread)
        per_thr = await clone_persistent(remote_thread, self.sock_path)
        await per_thr.reconnect(remote_thread)
        await assert_thread_works(self, per_thr)
        await per_thr.exit(0)

    async def test_ssh_new(self) -> None:
        "Start the persistent thread from one ssh thread, then reconnect to it from a new ssh thread."
        host = await make_local_ssh(self.thread)
        local_child, remote_thread = await host.ssh(self.thread)
        per_thr = await clone_persistent(remote_thread, self.sock_path)

        local_child, remote_thread = await host.ssh(self.thread)
        await per_thr.reconnect(remote_thread)
        await assert_thread_works(self, per_thr)

        await per_thr.exit(0)

    async def test_no_make_persistent(self) -> None:
        pidns_thr = await self.thread.clone(CLONE.NEWUSER|CLONE.NEWPID)
        sacr_thr = await pidns_thr.clone()
        await sacr_thr.task.setpgid()
        per_thr = await clone_persistent(sacr_thr, self.sock_path)
        # kill sacr_thr's process group to kill per_thr too
        await sacr_thr.process.killpg(SIG.KILL)
        # the persistent thread is dead, we can't reconnect to it
        with self.assertRaises(BaseException): # type: ignore
            await per_thr.reconnect(self.thread)

    async def test_make_persistent(self) -> None:
        # use a pidns so that the persistent task will be killed after all
        pidns_thr = await self.thread.clone(CLONE.NEWUSER|CLONE.NEWPID)
        sacr_thr = await pidns_thr.clone()
        await sacr_thr.task.setpgid()
        per_thr = await clone_persistent(sacr_thr, self.sock_path)
        # make the persistent thread, actually persistent.
        await per_thr.make_persistent()
        # kill sacr_thr's process group
        await sacr_thr.process.killpg(SIG.KILL)
        # the persistent thread is still around!
        await per_thr.reconnect(self.thread)
        await assert_thread_works(self, per_thr)
        await per_thr.exit(0)



from cffi import FFI
import os
import pathlib
import shutil

ffibuilder = FFI()
ffibuilder.set_source_pkgconfig(
    "rsyscall._raw", ["rsyscall"], """
#include <asm/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <linux/capability.h>
#include <linux/if_tun.h>
#include <linux/netlink.h>
#include <linux/futex.h>
#include <linux/fuse.h>
#include <linux/memfd.h>
#include <net/if.h>
#include <netinet/ip.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <poll.h>
#include <rsyscall.h>
#include <sched.h>
#include <setjmp.h>
#include <string.h>
#include <sys/epoll.h>
#include <sys/eventfd.h>
#include <sys/timerfd.h>
#include <sys/inotify.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/mount.h>
#include <sys/prctl.h>
#include <sys/ptrace.h>
#include <sys/resource.h>
#include <sys/signal.h>
#include <sys/signalfd.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <sys/uio.h>
#include <syscall.h>
#include <unistd.h>
#include <limits.h>

struct linux_dirent64 {
    ino64_t        d_ino;    /* 64-bit inode number */
    off64_t        d_off;    /* 64-bit offset to next structure */
    unsigned short d_reclen; /* Size of this dirent */
    unsigned char  d_type;   /* File type */
    char           d_name[]; /* Filename (null-terminated) */
};

// the double underscores are hard to use from Python,
// since they will be replaced with the class name
#define _WNOTHREAD __WNOTHREAD
#define _WCLONE __WCLONE
#define _WALL __WALL
#define _ss_padding __ss_padding

// glibc hides the real O_LARGEFILE from us - it lies to us! evil!
#undef O_LARGEFILE
#define O_LARGEFILE 0100000

#define SA_RESTORER 0x04000000
typedef void (*sighandler_t)(int);
typedef void (*sigrestore_t)(void);
// we assume NSIGNALS == 64, so we don't need any more than this
struct kernel_sigset {
    unsigned long int val;
};
struct kernel_sigaction {
	sighandler_t ksa_handler;
	unsigned long ksa_flags;
	sigrestore_t ksa_restorer;
	struct kernel_sigset ksa_mask;
};
struct fdpair {
    int first;
    int second;
};
struct futex_node {
  struct robust_list list;
  uint32_t futex;
};

// there are some buggy headers on older systems which have a negative value for EPOLLET :(
#undef EPOLLET
#define EPOLLET 0x80000000
""")
ffibuilder.cdef("""
typedef union epoll_data {
    uint64_t u64;
} epoll_data_t;
""")
ffibuilder.cdef("""
struct epoll_event {
  uint32_t     events;
  epoll_data_t data;
};
""", packed=True)
ffibuilder.cdef("""
struct fdpair {
    int first;
    int second;
};

typedef signed... time_t;

struct timespec {
    time_t tv_sec;                /* Seconds */
    long   tv_nsec;               /* Nanoseconds */
};

int epoll_wait(int epfd, struct epoll_event *events, int maxevents, int timeout);
int epoll_create1(int flags);
int epoll_ctl(int epfd, int op, int fd, struct epoll_event *event);

#define EPOLL_CTL_ADD ...
#define EPOLL_CTL_MOD ...
#define EPOLL_CTL_DEL ...

#define AT_FDCWD ...
#define AT_EMPTY_PATH ...
#define AT_SYMLINK_NOFOLLOW ...
#define AT_SYMLINK_FOLLOW ...
#define AT_REMOVEDIR ...

int unlinkat(int dirfd, const char *pathname, int flags);
int linkat(int olddirfd, const char *oldpath, int newdirfd, const char *newpath, int flags);
long rsyscall_raw_syscall(long arg1, long arg2, long arg3, long arg4, long arg5, long arg6, long sys);

#define EPOLL_CLOEXEC ...

typedef unsigned... ino64_t;
typedef signed... off64_t;

#define DT_BLK ... // This is a block device.
#define DT_CHR ... // This is a character device.
#define DT_DIR ... // This is a directory.
#define DT_FIFO ... // This is a named pipe (FIFO).
#define DT_LNK ... // This is a symbolic link.
#define DT_REG ... // This is a regular file.
#define DT_SOCK ... // This is a UNIX domain socket.
#define DT_UNKNOWN ... // The file type is unknown.

struct linux_dirent64 {
    ino64_t        d_ino;    /* 64-bit inode number */
    off64_t        d_off;    /* 64-bit offset to next structure */
    unsigned short d_reclen; /* Size of this dirent */
    unsigned char  d_type;   /* File type */
    char           d_name[]; /* Filename (null-terminated) */
};

// needed to determine true length of the null-terminated filenames, which are null-padded
size_t strlen(const char *s);

int faccessat(int dirfd, const char *pathname, int mode, int flags);

// signal stuff
#define SYS_waitid ...
#define SYS_kill ...

#define WEXITED ...
#define WSTOPPED ...
#define WCONTINUED ...
#define WNOHANG ...
#define WNOWAIT ...
#define _WCLONE ...
#define _WALL ...
#define _WNOTHREAD ...
#define P_PID ...
#define P_PGID ...
#define P_ALL ...
#define CLD_EXITED ... // child called _exit(2)
#define CLD_KILLED ... // child killed by signal
#define CLD_DUMPED ... // child killed by signal, and dumped core
#define CLD_STOPPED ... // child stopped by signal
#define CLD_TRAPPED ... // traced child has trapped
#define CLD_CONTINUED ... // child continued by SIGCONT
typedef int... pid_t;
typedef unsigned... uid_t;
typedef struct siginfo {
    int      si_signo;
    int      si_code;      /* Signal code */
    pid_t    si_pid;       /* Sending process ID */
    uid_t    si_uid;       /* Real user ID of sending process */
    int      si_status;    /* Exit value or signal */
    ...;
} siginfo_t;

#define SA_NOCLDSTOP ...
#define SA_NOCLDWAIT ...
#define SA_NODEFER ...
#define SA_ONSTACK ...
#define SA_RESETHAND ...
#define SA_RESTART ...
#define SA_SIGINFO ...
#define SA_RESTORER ...

typedef void (*sighandler_t)(int);
typedef void (*sigrestore_t)(void);
struct kernel_sigset {
    unsigned long int val;
};
struct kernel_sigaction {
	sighandler_t ksa_handler;
	unsigned long ksa_flags;
	sigrestore_t ksa_restorer;
	struct kernel_sigset ksa_mask;
};
#define SYS_rt_sigaction ...

#define SYS_signalfd4 ...

#define SFD_NONBLOCK ...
#define SFD_CLOEXEC ...

#define SYS_rt_sigprocmask ...

#define SIG_BLOCK ...
#define SIG_UNBLOCK ...
#define SIG_SETMASK ...

#define SIGSTKFLT ...

struct signalfd_siginfo {
    uint32_t ssi_signo;    /* Signal number */
    int32_t  ssi_errno;    /* Error number (unused) */
    int32_t  ssi_code;     /* Signal code */
    uint32_t ssi_pid;      /* PID of sender */
    uint32_t ssi_uid;      /* Real UID of sender */
    int32_t  ssi_fd;       /* File descriptor (SIGIO) */
    uint32_t ssi_tid;      /* Kernel timer ID (POSIX timers)
    uint32_t ssi_band;     /* Band event (SIGIO) */
    uint32_t ssi_overrun;  /* POSIX timer overrun count */
    uint32_t ssi_trapno;   /* Trap number that caused signal */
    int32_t  ssi_status;   /* Exit status or signal (SIGCHLD) */
    int32_t  ssi_int;      /* Integer sent by sigqueue(3) */
    uint64_t ssi_ptr;      /* Pointer sent by sigqueue(3) */
    uint64_t ssi_utime;    /* User CPU time consumed (SIGCHLD) */
    uint64_t ssi_stime;    /* System CPU time consumed
                              (SIGCHLD) */
    ...;
};

#define SYS_splice ...

// preadv2
#define SYS_preadv2 ...
#define SYS_pwritev2 ...

#define RWF_DSYNC ...
#define RWF_HIPRI ...
#define RWF_SYNC ...

// fs syscalls
#define SYS_openat ...
#define SYS_read ...
#define SYS_write ...
#define SYS_pread64 ...
#define SYS_pwrite64 ...
#define SYS_close ...
#define SYS_dup3 ...
#define SYS_pipe2 ...
#define SYS_ftruncate ...
#define SYS_ioctl ...
#define SYS_chmod ...
#define SYS_fchmod ...
#define SYS_fchmodat ...

#define SYS_chdir ...
#define SYS_fchdir ...
#define SYS_chroot ...

#define SYS_mkdirat ...
#define SYS_getdents64 ...
#define SYS_unlinkat ...
#define SYS_linkat ...
#define SYS_renameat2 ...

#define RENAME_EXCHANGE ...
#define RENAME_NOREPLACE ...
#define RENAME_WHITEOUT ...

#define SYS_symlinkat ...
#define SYS_readlinkat ...

#define O_RDONLY ...
#define O_WRONLY ...
#define O_RDWR ...
#define O_CREAT ...
#define O_EXCL ...
#define O_NOCTTY ...
#define O_TRUNC ...
#define O_APPEND ...
#define O_NONBLOCK ...
#define O_DSYNC ...
#define O_DIRECT ...
#define O_LARGEFILE ...
#define O_DIRECTORY ...
#define O_NOFOLLOW ...
#define O_NOATIME ...
#define O_CLOEXEC ...
#define O_SYNC ...
#define O_PATH ...
#define O_TMPFILE ...

#define SYS_lseek ...

#define SEEK_SET ...
#define SEEK_CUR ...
#define SEEK_END ...
#define SEEK_DATA ...
#define SEEK_HOLE ...

#define SYS_faccessat ...

#define R_OK ...
#define W_OK ...
#define X_OK ...
#define F_OK ...

// stat
#define SYS_fstat ...

#define S_IFMT   ...
#define S_IFSOCK ...
#define S_IFLNK	 ...
#define S_IFREG  ...
#define S_IFBLK  ...
#define S_IFDIR  ...
#define S_IFCHR  ...
#define S_IFIFO  ...
#define S_ISUID  ...
#define S_ISGID  ...
#define S_ISVTX  ...

#define S_IRWXU ...
#define S_IRUSR ...
#define S_IWUSR ...
#define S_IXUSR ...

#define S_IRWXG ...
#define S_IRGRP ...
#define S_IWGRP ...
#define S_IXGRP ...

#define S_IRWXO ...
#define S_IROTH ...
#define S_IWOTH ...
#define S_IXOTH ...

struct stat {
	unsigned long	st_dev;
	unsigned long	st_ino;
	unsigned long	st_nlink;

	unsigned int		st_mode;
	unsigned int		st_uid;
	unsigned int		st_gid;
	unsigned int		__pad0;
	unsigned long	st_rdev;
	long		st_size;
	long		st_blksize;
	long		st_blocks;	/* Number 512-byte blocks allocated. */

	struct timespec	st_atim;
	struct timespec	st_mtim;
	struct timespec	st_ctim;
	...;
};

// prctl
#define SYS_prctl ...
#define PR_GET_NAME ...
#define PR_SET_NAME ...
#define PR_SET_PDEATHSIG ...

#define SYS_getpid ...
#define SYS_setsid ...

#define SYS_setpgid ...
#define SYS_getpgid ...

// capabilities

#define SYS_capset ...
#define SYS_capget ...

#define _LINUX_CAPABILITY_VERSION_3 ...

#define PR_CAP_AMBIENT ...
#define PR_CAP_AMBIENT_RAISE ...

#define CAP_CHOWN ...
#define CAP_DAC_OVERRIDE ...
#define CAP_DAC_READ_SEARCH ...
#define CAP_FOWNER ...
#define CAP_FSETID ...
#define CAP_KILL ...
#define CAP_SETGID ...
#define CAP_SETUID ...
#define CAP_SETPCAP ...
#define CAP_LINUX_IMMUTABLE ...
#define CAP_NET_BIND_SERVICE ...
#define CAP_NET_BROADCAST ...
#define CAP_NET_ADMIN ...
#define CAP_NET_RAW ...
#define CAP_IPC_LOCK ...
#define CAP_IPC_OWNER ...
#define CAP_SYS_MODULE ...
#define CAP_SYS_RAWIO ...
#define CAP_SYS_CHROOT ...
#define CAP_SYS_PTRACE ...
#define CAP_SYS_PACCT ...
#define CAP_SYS_ADMIN ...
#define CAP_SYS_BOOT ...
#define CAP_SYS_NICE ...
#define CAP_SYS_RESOURCE ...
#define CAP_SYS_TIME ...
#define CAP_SYS_TTY_CONFIG ...
#define CAP_MKNOD ...
#define CAP_LEASE ...
#define CAP_AUDIT_WRITE ...
#define CAP_AUDIT_CONTROL ...
#define CAP_SETFCAP ...
#define CAP_MAC_OVERRIDE ...
#define CAP_MAC_ADMIN ...
#define CAP_SYSLOG ...
#define CAP_WAKE_ALARM ...
#define CAP_BLOCK_SUSPEND ...
#define CAP_AUDIT_READ ...
#define CAP_PERFMON ...
#define CAP_BPF ...
#define CAP_CHECKPOINT_RESTORE ...

struct __user_cap_data_struct {
    uint32_t effective;
    uint32_t permitted;
    uint32_t inheritable;
};
struct __user_cap_header_struct {
    uint32_t version;
    int pid;
};

// epoll stuff
#define SYS_epoll_ctl ...
#define SYS_epoll_wait ...
#define SYS_epoll_create1 ...

#define EPOLLIN ...
#define EPOLLOUT ...
#define EPOLLRDHUP ...
#define EPOLLPRI ...
#define EPOLLERR ...
#define EPOLLHUP ...
#define EPOLLET ...

// poll stuff
#define SYS_poll ...

struct pollfd {
    int   fd;         /* file descriptor */
    short events;     /* requested events */
    short revents;    /* returned events */
};

#define POLLIN ...
#define POLLHUP ...
#define POLLERR ...
#define POLLNVAL ...

//// eventfd
#define SYS_eventfd2 ...

#define EFD_CLOEXEC ...
#define EFD_NONBLOCK ...
#define EFD_SEMAPHORE ...

//// timerfd
#define SYS_timerfd_create ...
#define SYS_timerfd_settime ...
#define SYS_timerfd_gettime ...

#define CLOCK_REALTIME ...
#define CLOCK_MONOTONIC ...
#define CLOCK_BOOTTIME ...
#define CLOCK_REALTIME_ALARM ...
#define CLOCK_BOOTTIME_ALARM ...

#define TFD_NONBLOCK ...
#define TFD_CLOEXEC ...

#define TFD_TIMER_ABSTIME ...
#define TFD_TIMER_CANCEL_ON_SET ...

struct itimerspec {
    struct timespec it_interval;  /* Interval for periodic timer */
    struct timespec it_value;     /* Initial expiration */
};

//// inotify
#define SYS_inotify_init1 ...
#define IN_NONBLOCK ...
#define IN_CLOEXEC ...

#define SYS_inotify_add_watch ...
#define SYS_inotify_rm_watch ...

struct inotify_event {
    int      wd;       /* Watch descriptor */
    uint32_t mask;     /* Mask describing event */
    uint32_t cookie;   /* Unique cookie associating related
                          events (for rename(2)) */
    uint32_t len;      /* Size of name field */
    char     name[];   /* Optional null-terminated name */
};

#define NAME_MAX ...

// events settable in both
#define IN_ACCESS ...
#define IN_ATTRIB ...
#define IN_CLOSE_WRITE ...
#define IN_CLOSE_NOWRITE ...
#define IN_CREATE ...
#define IN_DELETE ...
#define IN_DELETE_SELF ...
#define IN_MODIFY ...
#define IN_MOVE_SELF ...
#define IN_MOVED_FROM ...
#define IN_MOVED_TO ...
#define IN_OPEN ...
// additional options to inotify_add_watch
#define IN_DONT_FOLLOW ...
#define IN_EXCL_UNLINK ...
#define IN_MASK_ADD ...
#define IN_ONESHOT ...
#define IN_ONLYDIR ...
// additional bits returned in struct inotify_event 
#define IN_IGNORED ...
#define IN_ISDIR ...
#define IN_Q_OVERFLOW ...
#define IN_UNMOUNT ...

//// task stuff
#define SYS_clone ...
#define SYS_vfork ...
#define SYS_exit ...
#define SYS_exit_group ...
#define SYS_execve ...
#define SYS_execveat ...
#define SYS_unshare ...
#define SYS_setns ...
#define SYS_set_tid_address ...

#define SYS_getuid ...
#define SYS_getgid ...

#define CLONE_VFORK ...
#define CLONE_CHILD_CLEARTID ...
#define CLONE_PARENT ...

#define CLONE_VM ...
#define CLONE_SIGHAND ...
#define CLONE_IO ...
#define CLONE_SYSVSEM ...

#define CLONE_FILES ...
#define CLONE_FS ...
#define CLONE_NEWCGROUP ...
#define CLONE_NEWIPC ...
#define CLONE_NEWNET ...
#define CLONE_NEWNS ...
#define CLONE_NEWPID ...
#define CLONE_NEWUSER ...
#define CLONE_NEWUTS ...

// sched_{set,get}affinity stuff

#define SYS_sched_setaffinity ...
#define SYS_sched_getaffinity ...

typedef struct {
  unsigned long __bits[...];
} cpu_set_t;

// getpriority/setpriority

#define PRIO_PROCESS ...
#define PRIO_PGRP ...
#define PRIO_USER ...

#define SYS_getpriority ...
#define SYS_setpriority ...

// prlimit

#define RLIMIT_AS ...
#define RLIMIT_CORE ...
#define RLIMIT_CPU ...
#define RLIMIT_DATA ...
#define RLIMIT_FSIZE ...
#define RLIMIT_LOCKS ...
#define RLIMIT_MEMLOCK ...
#define RLIMIT_MSGQUEUE ...
#define RLIMIT_NICE ...
#define RLIMIT_NOFILE ...
#define RLIMIT_NPROC ...
#define RLIMIT_RSS ...
#define RLIMIT_RTPRIO ...
#define RLIMIT_RTTIME ...
#define RLIMIT_SIGPENDING ...
#define RLIMIT_STACK ...

struct rlimit {
  uint64_t rlim_cur;
  uint64_t rlim_max;
};

#define SYS_prlimit64 ...

// mount stuff
#define SYS_mount ...
#define MS_BIND ...
#define MS_DIRSYNC ...
#define MS_LAZYTIME ...
#define MS_MANDLOCK ...
#define MS_MOVE ...
#define MS_NODEV ...
#define MS_NOEXEC ...
#define MS_NOSUID ...
#define MS_RDONLY ...
#define MS_REC ...
#define MS_RELATIME ...
#define MS_REMOUNT ...
#define MS_SILENT ...
#define MS_SLAVE ...
#define MS_STRICTATIME ...
#define MS_SYNCHRONOUS ...
#define MS_UNBINDABLE ...

#define SYS_umount2 ...
#define MNT_FORCE ...
#define MNT_DETACH ...
#define MNT_EXPIRE ...
#define UMOUNT_NOFOLLOW ...

//// FUSE stuff
// INIT flags
#define FUSE_ASYNC_READ		...
#define FUSE_POSIX_LOCKS	...
#define FUSE_FILE_OPS		...
#define FUSE_ATOMIC_O_TRUNC	...
#define FUSE_EXPORT_SUPPORT	...
#define FUSE_BIG_WRITES		...
#define FUSE_DONT_MASK		...
#define FUSE_SPLICE_WRITE	...
#define FUSE_SPLICE_MOVE	...
#define FUSE_SPLICE_READ	...
#define FUSE_FLOCK_LOCKS	...
#define FUSE_HAS_IOCTL_DIR	...
#define FUSE_AUTO_INVAL_DATA	...
#define FUSE_DO_READDIRPLUS	...
#define FUSE_READDIRPLUS_AUTO	...
#define FUSE_ASYNC_DIO		...
#define FUSE_WRITEBACK_CACHE	...
#define FUSE_NO_OPEN_SUPPORT	...
#define FUSE_PARALLEL_DIROPS    ...
#define FUSE_HANDLE_KILLPRIV	...
#define FUSE_POSIX_ACL		...
#define FUSE_ABORT_ERROR	...

// WRITE flags
#define FUSE_WRITE_CACHE	...
#define FUSE_WRITE_LOCKOWNER	...

// RELEASE flags
#define FUSE_RELEASE_FLUSH	...
#define FUSE_RELEASE_FLOCK_UNLOCK	...


enum fuse_opcode {
	FUSE_LOOKUP	   = ...,
	FUSE_FORGET	   = ..., /* no reply */
	FUSE_GETATTR	   = ...,
	FUSE_SETATTR	   = ...,
	FUSE_READLINK	   = ...,
	FUSE_SYMLINK	   = ...,
	FUSE_MKNOD	   = ...,
	FUSE_MKDIR	   = ...,
	FUSE_UNLINK	   = ...,
	FUSE_RMDIR	   = ...,
	FUSE_RENAME	   = ...,
	FUSE_LINK	   = ...,
	FUSE_OPEN	   = ...,
	FUSE_READ	   = ...,
	FUSE_WRITE	   = ...,
	FUSE_STATFS	   = ...,
	FUSE_RELEASE       = ...,
	FUSE_FSYNC         = ...,
	FUSE_SETXATTR      = ...,
	FUSE_GETXATTR      = ...,
	FUSE_LISTXATTR     = ...,
	FUSE_REMOVEXATTR   = ...,
	FUSE_FLUSH         = ...,
	FUSE_INIT          = ...,
	FUSE_OPENDIR       = ...,
	FUSE_READDIR       = ...,
	FUSE_RELEASEDIR    = ...,
	FUSE_FSYNCDIR      = ...,
	FUSE_GETLK         = ...,
	FUSE_SETLK         = ...,
	FUSE_SETLKW        = ...,
	FUSE_ACCESS        = ...,
	FUSE_CREATE        = ...,
	FUSE_INTERRUPT     = ...,
	FUSE_BMAP          = ...,
	FUSE_DESTROY       = ...,
	FUSE_IOCTL         = ...,
	FUSE_POLL          = ...,
	FUSE_NOTIFY_REPLY  = ...,
	FUSE_BATCH_FORGET  = ...,
	FUSE_FALLOCATE     = ...,
	FUSE_READDIRPLUS   = ...,
	FUSE_RENAME2       = ...,
	FUSE_LSEEK         = ...,

	/* CUSE specific operations */
	CUSE_INIT          = 4096,
};

#define FUSE_MIN_READ_BUFFER ...

// flags for fuse_attr
#define FUSE_ATTR_SUBMOUNT ...

struct fuse_attr {
	uint64_t	ino;
	uint64_t	size;
	uint64_t	blocks;
	uint64_t	atime;
	uint64_t	mtime;
	uint64_t	ctime;
	uint32_t	atimensec;
	uint32_t	mtimensec;
	uint32_t	ctimensec;
	uint32_t	mode;
	uint32_t	nlink;
	uint32_t	uid;
	uint32_t	gid;
	uint32_t	rdev;
	uint32_t	blksize;
	uint32_t	flags;
};

struct fuse_in_header {
    uint32_t len;       /* Total length of the data,
                           including this header */
    uint32_t opcode;    /* The kind of operation (see below) */
    uint64_t unique;    /* A unique identifier for this request */
    uint64_t nodeid;    /* ID of the filesystem object
                           being operated on */
    uint32_t uid;       /* UID of the requesting process */
    uint32_t gid;       /* GID of the requesting process */
    uint32_t pid;       /* PID of the requesting process */
    uint32_t padding;
};

struct fuse_out_header {
    uint32_t len;       /* Total length of data written to
                           the file descriptor */
    int32_t  error;     /* Any error that occurred (0 if none) */
    uint64_t unique;    /* The value from the
                           corresponding request */
};

struct fuse_init_in {
	uint32_t	major;
	uint32_t	minor;
	uint32_t	max_readahead;
	uint32_t	flags;
};

struct fuse_init_out {
	uint32_t	major;
	uint32_t	minor;
	uint32_t	max_readahead;
	uint32_t	flags;
	uint16_t	max_background;
	uint16_t	congestion_threshold;
	uint32_t	max_write;
	uint32_t	time_gran;
        ...;
};

// open_flags
#define FUSE_OPEN_KILL_SUIDGID	...

struct fuse_open_in {
	uint32_t	flags;
	uint32_t	open_flags;
};

#define FOPEN_DIRECT_IO   ...
#define FOPEN_KEEP_CACHE  ...
#define FOPEN_NONSEEKABLE ...

struct fuse_open_out {
	uint64_t	fh;
	uint32_t	open_flags;
	uint32_t	padding;
};

struct fuse_entry_out {
	uint64_t	nodeid;		/* Inode ID */
	uint64_t	generation;	/* Inode generation: nodeid:gen must
					   be unique for the fs's lifetime */
	uint64_t	entry_valid;	/* Cache timeout for the name */
	uint64_t	attr_valid;	/* Cache timeout for the attributes */
	uint32_t	entry_valid_nsec;
	uint32_t	attr_valid_nsec;
	struct fuse_attr attr;
};

#define FUSE_READ_LOCKOWNER ...

struct fuse_read_in {
	uint64_t	fh;
	uint64_t	offset;
	uint32_t	size;
	uint32_t	read_flags;
	uint64_t	lock_owner;
	uint32_t	flags;
	uint32_t	padding;
};

#define FUSE_GETATTR_FH ...

struct fuse_getattr_in {
	uint32_t	getattr_flags;
	uint32_t	dummy;
	uint64_t	fh;
};

struct fuse_attr_out {
	uint64_t	attr_valid;	/* Cache timeout for the attributes */
	uint32_t	attr_valid_nsec;
	uint32_t	dummy;
	struct fuse_attr attr;
};

struct fuse_dirent {
	uint64_t	ino;
	uint64_t	off;
	uint32_t	namelen;
	uint32_t	type;
	char name[];
};

struct fuse_direntplus {
	struct fuse_entry_out entry_out;
	struct fuse_dirent dirent;
};

struct fuse_flush_in {
	uint64_t	fh;
	uint32_t	unused;
	uint32_t	padding;
	uint64_t	lock_owner;
};

struct fuse_release_in {
	uint64_t	fh;
	uint32_t	flags;
	uint32_t	release_flags;
	uint64_t	lock_owner;
};

struct fuse_getxattr_in {
	uint32_t	size;
	uint32_t	padding;
};

struct fuse_getxattr_out {
	uint32_t	size;
	uint32_t	padding;
};

// socket stuff
#define SYS_socket ...
#define SYS_socketpair ...
#define SYS_bind ...
#define SYS_listen ...
#define SYS_accept4 ...
#define SYS_connect ...
#define SYS_getsockname ...
#define SYS_getpeername ...
#define SYS_shutdown ...

#define SHUT_RD ...
#define SHUT_WR ...
#define SHUT_RDWR ...

#define SOCK_NONBLOCK ...
#define SOCK_CLOEXEC ...

#define SOCK_DGRAM ...
#define SOCK_STREAM ...
#define SOCK_SEQPACKET ...
#define SOCK_RAW ...

typedef unsigned... sa_family_t;

#define AF_UNIX ...
struct sockaddr_un {
    sa_family_t sun_family;               /* AF_UNIX */
    char        sun_path[108];            /* pathname */
};

#define AF_INET ...
typedef unsigned... in_port_t;
struct sockaddr_in {
    sa_family_t    sin_family; /* address family: AF_INET */
    in_port_t      sin_port;   /* port in network byte order */
    struct in_addr sin_addr;   /* internet address */
    ...;
};

#define AF_INET6 ...
struct sockaddr_in6 {
    sa_family_t     sin6_family; /* address family: AF_INET6 */
    in_port_t       sin6_port;   /* port in network byte order */
    uint32_t        sin6_flowinfo; /* IPv6 flow information */
    struct in6_addr sin6_addr;     /* IPv6 address */
    uint32_t        sin6_scope_id; /* Scope ID (new in 2.4) */
};

struct in6_addr {
    unsigned char   s6_addr[16];
    ...;
};

struct sockaddr {
    sa_family_t    sa_family;
    char sa_data[14];
};

struct sockaddr_storage {
    sa_family_t    ss_family;
    char _ss_padding[...];
};

/* TCP setsockopt options, from netinet/tcp.h */

#define TCP_NODELAY              ... /* Don't delay send to coalesce packets  */
#define TCP_MAXSEG               ... /* Set maximum segment size  */
#define TCP_CORK                 ... /* Control sending of partial frames  */
#define TCP_KEEPIDLE             ... /* Start keeplives after this period */
#define TCP_KEEPINTVL            ... /* Interval between keepalives */
#define TCP_KEEPCNT              ... /* Number of keepalives before death */
#define TCP_SYNCNT               ... /* Number of SYN retransmits */
#define TCP_LINGER2              ... /* Life time of orphaned FIN-WAIT-2 state */
#define TCP_DEFER_ACCEPT         ... /* Wake up listener only when data arrive */
#define TCP_WINDOW_CLAMP         ... /* Bound advertised window */
#define TCP_INFO                 ... /* Information about this connection. */
#define TCP_QUICKACK             ... /* Bock/reenable quick ACKs.  */
#define TCP_CONGESTION           ... /* Congestion control algorithm.  */
#define TCP_MD5SIG               ... /* TCP MD5 Signature (RFC2385) */
#define TCP_COOKIE_TRANSACTIONS  ... /* TCP Cookie Transactions */
#define TCP_THIN_LINEAR_TIMEOUTS ... /* Use linear timeouts for thin streams*/
#define TCP_THIN_DUPACK          ... /* Fast retrans. after 1 dupack */
#define TCP_USER_TIMEOUT         ... /* How long for loss retry before timeout */
#define TCP_REPAIR               ... /* TCP sock is under repair right now */
#define TCP_REPAIR_QUEUE         ... /* Set TCP queue to repair */
#define TCP_QUEUE_SEQ            ... /* Set sequence number of repaired queue. */
#define TCP_REPAIR_OPTIONS       ... /* Repair TCP connection options */
#define TCP_FASTOPEN             ... /* Enable FastOpen on listeners */
#define TCP_TIMESTAMP            ... /* TCP time stamp */
#define TCP_NOTSENT_LOWAT        ... /* Limit number of unsent bytes in write queue.  */
#define TCP_CC_INFO              ... /* Get Congestion Control (optional) info.  */
#define TCP_SAVE_SYN             ... /* Record SYN headers for new connections.  */
#define TCP_SAVED_SYN            ... /* Get SYN headers recorded for connection.  */
#define TCP_REPAIR_WINDOW        ... /* Get/set window parameters. */
#define TCP_FASTOPEN_CONNECT     ... /* Attempt FastOpen with connect.  */
#define TCP_ULP                  ... /* Attach a ULP to a TCP connection.  */
#define TCP_MD5SIG_EXT           ... /* TCP MD5 Signature with extensions.  */
#define TCP_FASTOPEN_KEY         ... /* Set the key for Fast Open (cookie).  */
#define TCP_FASTOPEN_NO_COOKIE   ... /* Enable TFO without a TFO cookie.  */
#define TCP_ZEROCOPY_RECEIVE     ...
#define TCP_INQ                  ... /* Notify bytes available to read as a cmsg on read.  */
#define TCP_CM_INQ               ...
#define TCP_TX_DELAY             ... /* Delay outgoing packets by XX usec.  */


/* Internet address. */
struct in_addr {
    uint32_t       s_addr;     /* address in network byte order */
};

#define IFNAMSIZ ...

struct ifmap {
    ...;
};

// low level networking
struct ifreq {
    char ifr_name[...]; /* Interface name */
    union {
        struct sockaddr ifr_addr;
        struct sockaddr ifr_dstaddr;
        struct sockaddr ifr_broadaddr;
        struct sockaddr ifr_netmask;
        struct sockaddr ifr_hwaddr;
        short           ifr_flags;
        int             ifr_ifindex;
        int             ifr_metric;
        int             ifr_mtu;
        struct ifmap    ifr_map;
        char            ifr_slave[...];
        char            ifr_newname[...];
        char           *ifr_data;
    };
};

// Socket configuration ioctls
#define SIOCGIFNAME ... // get iface name
#define SIOCSIFLINK ... // set iface channel
#define SIOCGIFCONF ... // get iface list
#define SIOCGIFFLAGS ... // get flags
#define SIOCSIFFLAGS ... // set flags
#define SIOCGIFADDR ... // get PA address
#define SIOCSIFADDR ... // set PA address
#define SIOCGIFDSTADDR ... // get remote PA address
#define SIOCSIFDSTADDR ... // set remote PA address
#define SIOCGIFBRDADDR ... // get broadcast PA address
#define SIOCSIFBRDADDR ... // set broadcast PA address
#define SIOCGIFNETMASK ... // get network PA mask
#define SIOCSIFNETMASK ... // set network PA mask
#define SIOCGIFMETRIC ... // get metric
#define SIOCSIFMETRIC ... // set metric
#define SIOCGIFMEM ... // get memory address (BSD)
#define SIOCSIFMEM ... // set memory address (BSD)
#define SIOCGIFMTU ... // get MTU size
#define SIOCSIFMTU ... // set MTU size
#define SIOCSIFNAME ... // set interface name
#define SIOCSIFHWADDR ... // set hardware address
#define SIOCGIFENCAP ... // get/set encapsulations
#define SIOCSIFENCAP ...
#define SIOCGIFHWADDR ... // Get hardware address
#define SIOCGIFSLAVE ... // Driver slaving support
#define SIOCSIFSLAVE ...
#define SIOCADDMULTI ... // Multicast address lists
#define SIOCDELMULTI ...
#define SIOCGIFINDEX ... // name -> if_index mapping
#define SIOCSIFPFLAGS ... // set/get extended flags set
#define SIOCGIFPFLAGS ...
#define SIOCDIFADDR ... // delete PA address
#define SIOCSIFHWBROADCAST ... // set hardware broadcast addr
#define SIOCGIFCOUNT ... // get number of devices

// device flags, in ifr_flags with SIOC{G,S}IFFLAGS
#define IFF_UP ...          // Interface is running.
#define IFF_BROADCAST ...   // Valid broadcast address set.
#define IFF_DEBUG ...       // Internal debugging flag.
#define IFF_LOOPBACK ...    // Interface is a loopback interface.
#define IFF_POINTOPOINT ... // Interface is a point-to-point link.
#define IFF_RUNNING ...     // Resources allocated.
#define IFF_NOARP ...       // No arp protocol, L2 destination address not set.
#define IFF_PROMISC ...     // Interface is in promiscuous mode.
#define IFF_NOTRAILERS ...  // Avoid use of trailers.
#define IFF_ALLMULTI ...    // Receive all multicast packets.
#define IFF_MASTER ...      // Master of a load balancing bundle.
#define IFF_SLAVE ...       // Slave of a load balancing bundle.
#define IFF_MULTICAST ...   // Supports multicast
#define IFF_PORTSEL ...     // Is able to select media type via ifmap.
#define IFF_AUTOMEDIA ...   // Auto media selection active.
#define IFF_DYNAMIC ...     // The addresses are lost when the interface goes down.
// glibc doesn't have these in its headers for some reason
// #define IFF_LOWER_UP ...    // Driver signals L1 up (since Linux 2.6.17)
// #define IFF_DORMANT ...     // Driver signals dormant (since Linux 2.6.17)
// #define IFF_ECHO ...        // Echo sent packets (since Linux 2.6.25)

#define TUNSETIFF ...
// TUNSETIFF flags
#define IFF_TUN ...
#define IFF_TAP ...
#define IFF_NO_PI ...

// netlink

#define AF_NETLINK ...
#define NETLINK_ROUTE ...
struct sockaddr_nl {
    sa_family_t     nl_family;  /* AF_NETLINK */
    unsigned short  nl_pad;     /* Zero */
    pid_t           nl_pid;     /* Port ID */
    uint32_t        nl_groups;  /* Multicast groups mask */
};


// sockopt stuff
#define SYS_getsockopt ...
#define SYS_setsockopt ...

#define SOL_SOCKET ...
#define SO_ACCEPTCONN ...
#define SO_ATTACH_FILTER ...
#define SO_ATTACH_BPF ...
#define SO_ATTACH_REUSEPORT_CBPF ...
#define SO_ATTACH_REUSEPORT_EBPF ...
#define SO_BINDTODEVICE ...
#define SO_BROADCAST ...
#define SO_BSDCOMPAT ...
#define SO_DEBUG ...
#define SO_DETACH_FILTER ...
#define SO_DETACH_BPF ...
#define SO_DOMAIN ...
#define SO_ERROR ...
#define SO_DONTROUTE ...
#define SO_INCOMING_CPU ...
#define SO_KEEPALIVE ...
#define SO_LINGER ...
#define SO_LOCK_FILTER ...
#define SO_MARK ...
#define SO_OOBINLINE ...
#define SO_PASSCRED ...
#define SO_PEEK_OFF ...
#define SO_PEERCRED ...
#define SO_PRIORITY ...
#define SO_PROTOCOL ...
#define SO_RCVBUF ...
#define SO_RCVBUFFORCE ...
#define SO_RCVLOWAT ...
#define SO_SNDLOWAT ...
#define SO_RCVTIMEO ...
#define SO_SNDTIMEO ...
#define SO_REUSEADDR ...
#define SO_REUSEPORT ...
#define SO_RXQ_OVFL ...
#define SO_SNDBUF ...
#define SO_SNDBUFFORCE ...
#define SO_TIMESTAMP ...
#define SO_TYPE ...
#define SO_BUSY_POLL ...

#define SOL_IP ...

#define IPPROTO_ICMPV6 ...

#define IP_RECVERR ...
#define IP_PKTINFO ...
#define IP_MULTICAST_TTL ...
#define IP_MTU_DISCOVER ...
#define IP_PMTUDISC_DONT ...

// fcntl stuff
#define SYS_fcntl ...

#define F_SETFD ...
#define F_GETFD ...
#define F_SETFL ...
#define F_ADD_SEALS ...
#define F_GET_SEALS ...

#define F_SEAL_SEAL ...
#define F_SEAL_SHRINK ...
#define F_SEAL_GROW ...
#define F_SEAL_WRITE ...
#define F_SEAL_FUTURE_WRITE ...

#define FD_CLOEXEC ...

// mmap stuff
#define SYS_mmap ...
#define SYS_munmap ...
#define SYS_memfd_create ...

// memfd stuff, from sys/mman.h and linux/memfd.h
#define MFD_CLOEXEC ...
#define MFD_ALLOW_SEALING ...
#define MFD_HUGETLB ...
#define MFD_HUGE_2MB ...
#define MFD_HUGE_1GB ...

#define PROT_EXEC ...
#define PROT_READ ...
#define PROT_WRITE ...
#define PROT_NONE ...

#define MAP_SHARED ...
#define MAP_ANONYMOUS ...
#define MAP_PRIVATE ...
#define MAP_POPULATE ...
#define MAP_GROWSDOWN ...
#define MAP_STACK ...

void *memcpy(void *dest, const void *src, size_t n);
// we need these as function pointers, we aren't calling them from Python
int (*const rsyscall_persistent_server)(int infd, int outfd, const int listensock);
int (*const rsyscall_server)(const int infd, const int outfd);
void (*const rsyscall_futex_helper)(void *futex_addr);
void (*const rsyscall_trampoline)(void);

struct rsyscall_trampoline_stack {
    int64_t rdi;
    int64_t rsi;
    int64_t rdx;
    int64_t rcx;
    int64_t r8;
    int64_t r9;
    void* function;
};

struct rsyscall_syscall {
    int64_t sys;
    int64_t args[6];
};
struct rsyscall_symbol_table {
    void* rsyscall_server;
    void* rsyscall_persistent_server;
    void* rsyscall_futex_helper;
    void* rsyscall_trampoline;
};
struct rsyscall_bootstrap {
    struct rsyscall_symbol_table symbols;
    pid_t pid;
    int listening_sock;
    int syscall_sock;
    int data_sock;
    size_t envp_count;
};
struct rsyscall_stdin_bootstrap {
    struct rsyscall_symbol_table symbols;
    pid_t pid;
    int syscall_fd;
    int data_fd;
    int futex_memfd;
    int connecting_fd;
    size_t envp_count;
};
struct rsyscall_unix_stub {
    struct rsyscall_symbol_table symbols;
    pid_t pid;
    int syscall_fd;
    int data_fd;
    int futex_memfd;
    int connecting_fd;
    size_t argc;
    size_t envp_count;
    uint64_t sigmask;
};

// sockets
#define SYS_sendto ...
#define SYS_recvfrom ...

#define SYS_sendmsg ...
#define SYS_recvmsg ...

#define SCM_RIGHTS ...

// send flags
#define MSG_CONFIRM ...
#define MSG_DONTROUTE ...
#define MSG_EOR ...
#define MSG_MORE ...
#define MSG_NOSIGNAL ...
// recv flags
#define MSG_CMSG_CLOEXEC ...
#define MSG_ERRQUEUE ...
#define MSG_PEEK ...
#define MSG_TRUNC ...
#define MSG_WAITALL ...
// both
#define MSG_DONTWAIT ...
#define MSG_OOB ...
// recvmsg returned msg_flags field
#define MSG_CTRUNC ...

struct iovec {
    void *iov_base;	/* Pointer to data.  */
    size_t iov_len;	/* Length of data.  */
};

typedef unsigned long int socklen_t;
struct msghdr {
    void *msg_name;		/* Address to send to/receive from.  */
    int msg_namelen;     	/* Length of address data.  */
  
    struct iovec *msg_iov;	/* Vector of data to send/receive into.  */
    unsigned long int msg_iovlen;		/* Number of elements in the vector.  */
  
    void *msg_control;		/* Ancillary data (eg BSD filedesc passing). */
    unsigned long int msg_controllen;	/* Ancillary data buffer length.  */
  
    int msg_flags;		/* Flags in received message.  */
};

struct cmsghdr {
    unsigned long int cmsg_len;		/* Length of data in cmsg_data plus length
				   of cmsghdr structure.  */
    int cmsg_level;		/* Originating protocol.  */
    int cmsg_type;		/* Protocol specific type.  */
    ...;
};

//// ugh, have to take this over to get notification of thread exec
#define SYS_set_robust_list ...

// see kernel source for documentation of robust_lists.
// this is our custom robust_list + futex structure
struct futex_node {
  struct robust_list list;
  uint32_t futex;
};

struct robust_list {
  struct robust_list *next;
};

struct robust_list_head {
  struct robust_list list;
  unsigned long futex_offset;
  struct robust_list *list_op_pending;
};

#define FUTEX_WAITERS ...
#define FUTEX_TID_MASK ...

""")



"`#include <linux/futex.h>`"
from __future__ import annotations
import typing as t
from rsyscall._raw import ffi, lib # type: ignore
import enum
from dataclasses import dataclass
from rsyscall.struct import Struct
from rsyscall.handle import Pointer, WrittenPointer

FUTEX_WAITERS: int = lib.FUTEX_WAITERS
FUTEX_TID_MASK: int = lib.FUTEX_TID_MASK

@dataclass
class FutexNode(Struct):
    # this is our bundle of struct robust_list with a futex.  since it's tricky to handle the
    # reference management of taking a reference to just one field in a structure (the futex, in
    # cases where we don't care about the robust list), we always deal in the entire FutexNode
    # structure whenever we talk about futexes. that's a bit of overhead but we barely use futexes,
    # so it's fine.
    next: t.Optional[Pointer[FutexNode]]
    futex: int

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct futex_node*', {
            # technically we're supposed to have a pointer to the first node in the robust list to
            # indicate the end.  but that's tricky to do. so instead let's just use a NULL pointer;
            # the kernel will EFAULT when it hits the end. make sure not to map 0, or we'll
            # break. https://imgflip.com/i/2zwysg
            'list': (ffi.cast('struct robust_list*', int(self.next.near)) if self.next else ffi.NULL,),
            'futex': self.futex,
        })
        return bytes(ffi.buffer(struct))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct futex_node')

@dataclass
class RobustListHead(Struct):
    first: WrittenPointer[FutexNode]

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct robust_list_head*', {
            'list': (ffi.cast('struct robust_list*', int(self.first.near)),),
            'futex_offset': ffi.offsetof('struct futex_node', 'futex'),
            'list_op_pending': ffi.NULL,
        })
        return bytes(ffi.buffer(struct))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct robust_list_head')

#### Classes ####
import rsyscall.far
from rsyscall.handle.pointer import WrittenPointer

class FutexTask(rsyscall.far.Task):
    async def set_robust_list(self, head: WrittenPointer[RobustListHead]) -> None:
        with head.borrow(self):
            await _set_robust_list(self.sysif, head.near, head.size())

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _set_robust_list(sysif: SyscallInterface, head: near.Address, len: int) -> None:
    await sysif.syscall(SYS.set_robust_list, head, len)



"""Async equivalent of compile, which takes an AST and returns an awaitable

"""
from dataclasses import dataclass
import ast
import builtins
import types
import inspect
import typing as t

@dataclass
class _InternalResult(Exception):
    is_expression: bool
    value: t.Any

__result_exception__ = _InternalResult

def compile_to_awaitable(astob: ast.Interactive,
                         global_vars: t.Dict[str, t.Any]) -> t.Awaitable:
    """Compile this AST, which may contain await statements, to an awaitable.

    - If the AST calls return, then a value is returned from the awaitable.
    - If the AST raises an exception, then the awaitable raises that exception.
    - If the AST neither returns a value nor raises an exception, then __result_exception__ is
      raised.
    - If the last statement in the AST is an expression, then on the __result_exception__
      exception, is_expression is set and value contains the value of the expression.
    - If the last statement in the AST is not an expression, then on the __result_exception__
      exception, is_expression is False and value contains None.

    """
    wrapper_name = "__toplevel__"
    # we rely on the user not messing with __builtins__ in the REPL; that's something you
    # really aren't supposed to do, so I think that's fine.
    wrapper = ast.parse(f"""
async def {wrapper_name}():
    try:
        pass
    finally:
        __builtins__.locals()
""", filename="<internal_wrapper>", mode="single")
    try_block = wrapper.body[0].body[0] # type: ignore
    try_block.body = astob.body
    if isinstance(try_block.body[-1], (ast.Expr, ast.Await)):
        # if the last statement in the AST is an expression, then have its value be
        # propagated up by throwing it from the __result_exception__ exception.
        wrapper_raise = ast.parse("raise __result_exception__(True, None)", filename="<internal_wrapper>", mode="single").body[0] # type: ignore
        wrapper_raise.exc.args[1] = try_block.body[-1].value # type: ignore
        try_block.body[-1] = wrapper_raise
    else:
        wrapper_raise = ast.parse("raise __result_exception__(False, None)", filename="<internal_wrapper>", mode="single").body[0] # type: ignore
        try_block.body.append(wrapper_raise)
    global_vars.update({
        '__builtins__': builtins,
        '__result_exception__': __result_exception__,
    })
    exec(compile(wrapper, '<input>', 'single'), global_vars)
    func = global_vars[wrapper_name]
    del global_vars[wrapper_name]
    # don't create a new local variable scope
    func.__code__ = func.__code__ .replace(co_flags=func.__code__.co_flags & ~inspect.CO_NEWLOCALS)
    return func()
